#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import stat
import datetime

import netius.common

BUFFER_SIZE = 16384
""" The size of the buffer that is going to be used when
sending the file to the client, this should not be neither
to big nor to small (as both situations would create problems) """

CAPABILITIES = (
    "PASV",
    "UTF8"
)
""" The sequence defining the complete set of capabilities
that are available under the current ftp server implementation """

PERMISSIONS = {
    7 : "rwx",
    6 : "rw-",
    5 : "r-x",
    4 : "r--",
    0 : "---"
}
""" Map that defines the association between the octal based
values for the permissions and the associated string values """

TYPES = {
    "A" : "ascii",
    "E" : "ebcdic",
    "I" : "binary",
    "L" : "local"
}
""" The map that associated the various type command arguments
with the more rich data mode transfer types """

class FTPConnection(netius.Connection):

    def __init__(self, base_path = "", host = "ftp.localhost", mode = "ascii", *args, **kwargs):
        netius.Connection.__init__(self, *args, **kwargs)
        self.parser = None
        self.base_path = os.path.abspath(base_path)
        self.host = host
        self.mode = mode
        self.data_server = None
        self.remaining = None
        self.cwd = "/"
        self.username = "anonymous"
        self.password = "anonymous"

    def open(self, *args, **kwargs):
        netius.Connection.open(self, *args, **kwargs)
        if not self.is_open(): return
        self.parser = netius.common.FTPParser(self)
        self.parser.bind("on_line", self.on_line)

    def close(self, *args, **kwargs):
        netius.Connection.close(self, *args, **kwargs)
        if not self.is_closed(): return
        if self.parser: self.parser.destroy()
        if self.data_server: self.data_server.close_ftp()
        file = hasattr(self, "file") and self.file
        if file: file.close()

    def parse(self, data):
        return self.parser.parse(data)

    def send_ftp(self, code, message = "", lines = (), simple = False, delay = True, callback = None):
        if lines: return self.send_ftp_lines(
            code,
            message = message,
            lines = lines,
            simple = simple,
            delay = delay,
            callback = callback
        )
        else: return self.send_ftp_base(
            code,
            message,
            delay,
            callback
        )

    def send_ftp_base(self, code, message = "", delay = True, callback = None):
        base = "%d %s" % (code, message)
        data = base + "\r\n"
        count = self.send(data, delay = delay, callback = callback)
        self.owner.debug(base)
        return count

    def send_ftp_lines(self, code, message = "", lines = (), simple = False, delay = True, callback = None):
        lines = list(lines)
        if not simple: lines.insert(0, message)
        body = lines[:-1]
        tail = lines[-1]
        base = "%d-%s" % (code, message) if simple else "%d %s" % (code, message)
        lines_s = [" %s" % line for line in body] if simple else\
            ["%d-%s" % (code, line) for line in body]
        lines_s.append("%d %s" % (code, tail))
        if simple: lines_s.insert(0, base)
        data = "\r\n".join(lines_s) + "\r\n"
        count = self.send(data, delay = delay, callback = callback)
        self.owner.debug(base)
        return count

    def ready(self):
        message = "%s FTP Server %s ready" % (self.host, netius.NAME)
        self.send_ftp(220, message)

    def ok(self):
        message = "ok"
        self.send_ftp(200, message)

    def not_ok(self):
        message = "not ok"
        self.send_ftp(500, message)

    def flush_ftp(self):
        if not self.remaining: return
        method = getattr(self, "flush_" + self.remaining)
        try: method()
        finally: self.remaining = None

    def data_ftp(self, data):
        self.file.write(data)

    def closed_ftp(self):
        has_file = hasattr(self, "file") and not self.file == None
        if not has_file: return
        self.file.close()
        self.file = None
        self.send_ftp(226, "file receive ok")

    def flush_list(self):
        self.send_ftp(150, "directory list sending")
        list_data = self._list()
        self.data_server.send_ftp(list_data, callback = self.on_flush_list)

    def flush_retr(self):
        self.send_ftp(150, "file sending")
        full_path = self._get_path(self.file_name)
        self.bytes_p = os.path.getsize(full_path)
        self.file = open(full_path, "rb")
        self._file_send(self)

    def flush_stor(self):
        self.send_ftp(150, "file receiving")

        full_path = self._get_path(self.file_name)
        self.file = open(full_path, "wb")

    def on_flush_list(self, connection):
        self._data_close()
        self.send_ftp(226, "directory send ok")

    def on_flush_retr(self, connection):
        self._data_close()
        self.send_ftp(226, "file send ok")

    def on_line(self, code, message, is_final = True):
        # "joins" the code and the message part of the message into the base
        # string and then uses this value to print some debug information
        base = "%s %s" % (code, message)
        self.owner.debug(base)

        # calls the proper top level owner based line information handler that
        # should ignore any usages as the connection will take care of the proper
        # handling for the current connection
        self.owner.on_line_ftp(self, code, message)

        # converts the provided code into a lower case value and then uses it
        # to create the problem name for the handler method to be used
        code_l = code.lower()
        method_n = "on_" + code_l

        # verifies if the method for the current code exists in case it
        # does not raises an exception indicating the problem with the
        # code that has just been received (probably erroneous)
        extists = hasattr(self, method_n)
        if not extists: raise netius.ParserError("Invalid code '%s'" % code)

        # retrieves the reference to the method that is going to be called
        # for the handling of the current line from the current instance and
        # then calls it with the provided message
        method = getattr(self, method_n)
        method(message)

    def on_user(self, message):
        self.username = message
        self.ok()

    def on_syst(self, message):
        self.send_ftp(215, message = "UNIX Type: L8 (%s)" % netius.VERSION)

    def on_feat(self, message):
        self.send_ftp(211, "features", lines = list(CAPABILITIES) + ["end"], simple = True)

    def on_opts(self, message):
        self.ok()

    def on_pwd(self, message):
        self.send_ftp(257, "\"%s\"" % self.cwd)

    def on_type(self, message):
        self.mode = TYPES.get("message", "ascii")
        self.ok()

    def on_pasv(self, message):
        data_server = self._data_open()
        port_h = (data_server.port & 0xff00) >> 8
        port_l = data_server.port & 0x00ff
        address = self.socket.getsockname()[0]
        address = address.replace(".", ",")
        address_s = "%s,%d,%d" % (address, port_h, port_l)
        self.send_ftp(227, "entered passive mode (%s)" % address_s)

    def on_port(self, message):
        self.ok()

    def on_dele(self, message):
        full_path = self._get_path(extra = message)
        try: os.remove(full_path)
        except Exception: self.not_ok()
        else: self.ok()

    def on_mkd(self, message):
        full_path = self._get_path(extra = message)
        try: os.makedirs(full_path)
        except Exception: self.not_ok()
        else: self.ok()

    def on_rmd(self, message):
        full_path = self._get_path(extra = message)
        try: os.rmdir(full_path)
        except Exception: self.not_ok()
        else: self.ok()

    def on_rnfr(self, message):
        self.source_path = self._get_path(extra = message)
        self.ok()

    def on_rnto(self, message):
        self.target_path = self._get_path(extra = message)
        try: os.rename(self.source_path, self.target_path)
        except Exception: self.not_ok()
        else: self.ok()
        finally: self.source_path = self.target_path = None

    def on_cdup(self, message):
        self.cwd = self.cwd.rsplit("/", 1)[0]
        if not self.cwd: self.cwd = "/"
        self.ok()

    def on_cwd(self, message):
        is_absolute = message.startswith("/")
        if is_absolute: cwd = message
        else: cwd = self.cwd + (message if self.cwd.endswith("/") else "/" + message)

        full_path = self._get_path(extra = message)
        is_dir = os.path.isdir(full_path)
        if not is_dir: self.send_ftp(550, "failed to change directory"); return

        self.cwd = cwd
        self.ok()

    def on_size(self, message):
        full_path = self._get_path(extra = message)
        if os.path.isdir(full_path): size = 0
        else: size = os.path.getsize(full_path)
        self.send_ftp(213, "%d" % size)

    def on_mdtm(self, message):
        full_path = self._get_path(extra = message)
        if os.path.isdir(full_path): modified = 0
        else: modified = os.path.getmtime(full_path)
        modified_d = datetime.datetime.utcfromtimestamp(modified)
        modified_s = modified_d.strftime("%Y%m%d%H%M%S")
        self.send_ftp(213, modified_s)

    def on_noop(self, message):
        self.ok()

    def on_quit(self, message):
        self.send_ftp(221, "exiting connection")

    def on_list(self, message):
        self.remaining = "list"
        self.data_server.flush_ftp()

    def on_retr(self, message):
        self.remaining = "retr"
        self.file_name = message
        self.data_server.flush_ftp()

    def on_stor(self, message):
        self.remaining = "stor"
        self.file_name = message
        self.data_server.flush_ftp()

    def _file_send(self, connection):
        file = self.file
        is_larger = BUFFER_SIZE > self.bytes_p
        buffer_s = self.bytes_p if is_larger else BUFFER_SIZE
        data = file.read(buffer_s)
        data_l = len(data) if data else 0
        self.bytes_p -= data_l
        is_final = not data or self.bytes_p == 0
        callback = self._file_finish if is_final else self._file_send
        self.data_server.send_ftp(data, callback = callback)

    def _file_finish(self, connection):
        self.file.close()
        self.file = None
        self.bytes_p = None
        self.on_flush_retr(connection)

    def _data_open(self):
        if self.data_server: self._data_close()
        self.data_server = FTPDataServer(self, self.owner)
        self.data_server.serve(
            host = self.owner.host,
            port = 0,
            load = False,
            start = False
        )
        return self.data_server

    def _data_close(self):
        if not self.data_server: return
        self.data_server.close_ftp()
        self.data_server = None

    def _list(self):
        # retrieves the current value for the relative path, this
        # should take into account the current working directory
        relative_path = self._get_path()

        # lists the directory for the current relative path, this
        # should get a list of files contained in it, in case there's
        # an error in such listing an empty string is returned
        try: entries = os.listdir(relative_path)
        except Exception: return ""

        # allocates space for the list that will hold the various lines
        # for the complete set of tiles in the directory
        lines = []

        # iterates over the complete set of entries in the current
        # working directory to create their respective listing line
        for entry in entries:
            file_path = os.path.join(relative_path, entry)
            try: mode = os.stat(file_path)
            except Exception: continue
            permissions = self._to_unix(mode)
            timestamp = mode.st_mtime
            date_time = datetime.datetime.utcfromtimestamp(timestamp)
            date_s = date_time.strftime("%b %d  %Y")
            line = "%s    1 %-8s %-8s %8lu %s %s\r\n" %\
                (permissions, "ftp", "ftp", mode.st_size, date_s, entry)
            lines.append(line)

        # returns the final list string result as the joining of the
        # various lines for each of the files (as expected)
        return "".join(lines)

    def _to_unix(self, mode):
        is_dir = "d" if stat.S_ISDIR(mode.st_mode) else "-"
        permissions = str(oct(mode.st_mode)[-3:])
        return is_dir + "".join([PERMISSIONS.get(int(item), item) for item in permissions])

    def _get_path(self, extra = None):
        # tries to decide on own to resolve the base and extra parts
        # of the path taking into account a possible absolute extra
        # value, the current working directory is only used in case
        # the provided extra value is not absolute
        is_absolute = extra.startswith("/") if extra else False
        base = extra[1:] if is_absolute else self.cwd[1:]
        extra = "" if is_absolute else extra or ""

        # gathers the current relative (full) path for the state using
        # the current working directory value and normalizing it, note
        # that in case an extra (relative) path is provided it's joined
        # with the base path to provide the "new" full path
        relative_path = os.path.join(self.base_path, base)
        relative_path = os.path.join(relative_path, extra)
        relative_path = os.path.abspath(relative_path)
        relative_path = os.path.normpath(relative_path)
        return relative_path

class FTPDataServer(netius.StreamServer):

    def __init__(self, connection, container, *args, **kwargs):
        netius.StreamServer.__init__(self, *args, **kwargs)
        self.connection = connection
        self.container = container
        self.accepted = None
        self.container.add_base(self)

    def on_connection_c(self, connection):
        netius.StreamServer.on_connection_c(self, connection)
        if self.accepted: connection.close(); return
        self.accepted = connection
        self.flush_ftp()

    def on_connection_d(self, connection):
        netius.StreamServer.on_connection_d(self, connection)
        self.connection.closed_ftp()

    def on_data(self, connection, data):
        netius.StreamServer.on_data(self, connection, data)
        self.connection.data_ftp(data)

    def send_ftp(self, data, delay = True, force = False, callback = None):
        if not self.accepted: raise netius.DataError("No connection accepted")
        return self.accepted.send(data, delay = delay, force = force, callback = callback)

    def flush_ftp(self):
        if not self.accepted: return
        self.connection.flush_ftp()

    def close_ftp(self):
        if self.accepted: self.accepted.close(); self.accepted = None
        self.cleanup()

class FTPServer(netius.ContainerServer):
    """
    Abstract ftp server implementation that handles authentication
    and file system based file serving.

    Note that the ftp server does not support multiple user handling
    and runs only as the user running the current process.

    :see: http://tools.ietf.org/html/rfc959
    """

    def __init__(self, base_path = "", auth_s = "dummy", *args, **kwargs):
        netius.ContainerServer.__init__(self, *args, **kwargs)
        self.base_path = base_path
        self.auth_s = auth_s

    def serve(self, host = "ftp.localhost", port = 21, *args, **kwargs):
        netius.ContainerServer.serve(self, port = port, *args, **kwargs)
        self.host = host

    def on_connection_c(self, connection):
        netius.ContainerServer.on_connection_c(self, connection)
        connection.ready()

    def on_data(self, connection, data):
        netius.ContainerServer.on_data(self, connection, data)
        connection.parse(data)

    def on_serve(self):
        netius.ContainerServer.on_serve(self)
        if self.env: self.base_path = self.get_env("BASE_PATH", self.base_path)
        if self.env: self.host = self.get_env("FTP_HOST", self.host)
        if self.env: self.auth_s = self.get_env("FTP_AUTH", self.auth_s)
        self.auth = self.get_auth(self.auth_s)
        self.info("Starting FTP server on '%s' using '%s' ..." % (self.host, self.auth_s))
        self.info("Defining '%s' as the root of the file server ..." % (self.base_path or "."))

    def build_connection(self, socket, address, ssl = False):
        return FTPConnection(
            owner = self,
            socket = socket,
            address = address,
            ssl = ssl,
            base_path = self.base_path,
            host = self.host
        )

    def on_line_ftp(self, connection, code, message):
        pass

if __name__ == "__main__":
    import logging
    server = FTPServer(level = logging.DEBUG)
    server.serve(env = True)
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import time
import zlib
import base64
import tempfile

import netius.common

from netius.common import PLAIN_ENCODING, CHUNKED_ENCODING,\
    GZIP_ENCODING, DEFLATE_ENCODING

Z_PARTIAL_FLUSH = 1
""" The zlib constant value representing the partial flush
of the current zlib stream, this value has to be defined
locally as it is not defines under the zlib module """

class HTTPProtocol(netius.StreamProtocol):
    """
    Implementation of the HTTP protocol to be used by a client
    of the HTTP implementation to send requests and receive
    responses.
    """

    BASE_HEADERS = {
        "user-agent" : netius.IDENTIFIER
    }
    """ The map containing the complete set of headers
    that are meant to be applied to all the requests """

    def __init__(
        self,
        method,
        url,
        params = None,
        headers = None,
        data = None,
        version = "HTTP/1.1",
        encoding = PLAIN_ENCODING,
        encodings = "gzip, deflate",
        safe = False,
        request = False,
        asynchronous = True,
        timeout = None,
        use_file = False,
        callback = None,
        on_init = None,
        on_open = None,
        on_close = None,
        on_headers = None,
        on_data = None,
        on_result = None,
        *args,
        **kwargs
    ):
        netius.StreamProtocol.__init__(self, *args, **kwargs)
        self.parser = None
        self.set(
            method,
            url,
            params = params,
            headers = headers,
            data = data,
            version = version,
            encoding = encoding,
            encodings = encodings,
            safe = safe,
            request = request,
            asynchronous = asynchronous,
            timeout = timeout,
            use_file = use_file,
            callback = callback,
            on_init = on_init,
            on_open = on_open,
            on_close = on_close,
            on_headers = on_headers,
            on_data = on_data,
            on_result = on_result
        )

    @classmethod
    def key_g(cls, url):
        parsed = netius.legacy.urlparse(url)
        ssl = parsed.scheme == "https"
        host = parsed.hostname
        port = parsed.port or (443 if ssl else 80)
        return (host, port, ssl)

    @classmethod
    def decode_gzip(cls, data):
        if not data: return data
        return zlib.decompress(data, zlib.MAX_WBITS | 16)

    @classmethod
    def decode_deflate(cls, data):
        if not data: return data
        try: return zlib.decompress(data)
        except Exception: return zlib.decompress(data, -zlib.MAX_WBITS)

    @classmethod
    def decode_zlib_file(
        cls,
        input,
        output,
        buffer_size = 16384,
        wbits = zlib.MAX_WBITS | 16
    ):
        decompressor = zlib.decompressobj(wbits)
        while True:
            data = input.read(buffer_size)
            if not data: break
            raw_data = decompressor.decompress(data)
            output.write(raw_data)
        raw_data = decompressor.flush()
        output.write(raw_data)
        return output

    @classmethod
    def decode_gzip_file(
        cls,
        input,
        output,
        buffer_size = 16384,
        wbits = zlib.MAX_WBITS | 16
    ):
        return cls.decode_zlib_file(
            input,
            output,
            buffer_size = buffer_size,
            wbits = wbits
        )

    @classmethod
    def decode_deflate_file(
        cls,
        input,
        output,
        buffer_size = 16384,
        wbits = -zlib.MAX_WBITS
    ):
        return cls.decode_zlib_file(
            input,
            output,
            buffer_size = buffer_size,
            wbits = wbits
        )

    @classmethod
    def set_request(cls, parser, buffer, request = None):
        if request == None: request = dict()
        headers = parser.get_headers()
        data = b"".join(buffer)
        encoding = headers.get("Content-Encoding", None)
        decoder = getattr(cls, "decode_%s" % encoding) if encoding else None
        if decoder and data: data = decoder(data)
        request["code"] = parser.code
        request["status"] = parser.status
        request["headers"] = headers
        request["data"] = data
        return request

    @classmethod
    def set_request_file(
        cls,
        parser,
        input,
        request = None,
        output = None,
        buffer_size = 16384
    ):
        # verifies if a request object has been passes to the current
        # method and if that's not the case creates a new one (as a map)
        if request == None: request = dict()

        # retrieves the complete set of headers and tries discover the
        # encoding of it and the associated decoder (if any)
        headers = parser.get_headers()
        encoding = headers.get("Content-Encoding", None)
        decoder = getattr(cls, "decode_%s_file" % encoding) if encoding else None

        # in case there's a decoder and an input (file) then runs the decoding
        # process setting the data as the resulting (decoded object)
        if decoder and input:
            if output == None: output = tempfile.NamedTemporaryFile(mode = "w+b")
            input.seek(0)
            try:
                data = decoder(
                    input,
                    output,
                    buffer_size = buffer_size
                )
            finally:
                input.close()

        # otherwise it's a simplified process (no decoding required) and the
        # data may be set directly as the input file
        else:
            data = input

        # seeks the data object to the initial position so that it
        # is set as ready to be read from a third party
        data.seek(0)

        # updates the structure of the request object/map so that it
        # contains a series of information on the request, including
        # the file contents (stored in a temporary file)
        request["code"] = parser.code
        request["status"] = parser.status
        request["headers"] = headers
        request["data"] = data

        # returns the request object that has just been populated
        # to the caller method so that it may be used to read the contents
        return request

    @classmethod
    def set_error(cls, error, message = None, request = None, force = False):
        if request == None: request = dict()
        if "error" in request and not force: return
        request["error"] = error
        request["message"] = message
        return request

    def open_c(self, *args, **kwargs):
        netius.StreamProtocol.open_c(self, *args, **kwargs)

        # creates a new HTTP parser instance and set the correct event
        # handlers so that the data parsing is properly handled
        self.parser = netius.common.HTTPParser(self, type = netius.common.RESPONSE)
        self.parser.bind("on_data", self._on_data)
        self.parser.bind("on_partial", self.on_partial)
        self.parser.bind("on_headers", self.on_headers)
        self.parser.bind("on_chunk", self.on_chunk)

    def close_c(self, *args, **kwargs):
        netius.StreamProtocol.close_c(self, *args, **kwargs)

        if self.parser: self.parser.destroy()
        if self.parsed: self.parsed = None
        if self.gzip: self._close_gzip(safe = True)
        if self.gzip_c: self.gzip_c = None

    def info_dict(self, full = False):
        info = netius.StreamProtocol.info_dict(self, full = full)
        info.update(
            version = self.version,
            method = self.method,
            encoding = self.encodings,
            url = self.url,
            parsed = self.parsed,
            host = self.host,
            port = self.port,
            path = self.path,
            headers = self.headers
        )
        if full: info.update(
            parser = self.parser.info_dict()
        )
        return info

    def connection_made(self, transport):
        netius.StreamProtocol.connection_made(self, transport)

        # performs the run request operation that should trigger
        # the process of sending the request to the server
        self.run_request()

    def loop_set(self, loop):
        netius.StreamProtocol.loop_set(self, loop)
        self.set_dynamic()

    def flush(self, force = False, callback = None):
        if self.current == DEFLATE_ENCODING:
            self._flush_gzip(force = force, callback = callback)
        elif self.current == GZIP_ENCODING:
            self._flush_gzip(force = force, callback = callback)
        elif self.current == CHUNKED_ENCODING:
            self._flush_chunked(force = force, callback = callback)
        elif self.current == PLAIN_ENCODING:
            self._flush_plain(force = force, callback = callback)

        self.current = self.encoding

    def send_base(
        self,
        data,
        stream = None,
        final = True,
        delay = True,
        force = False,
        callback = None
    ):
        data = netius.legacy.bytes(data) if data else data
        if self.current == PLAIN_ENCODING:
            return self.send_plain(
                data,
                stream = stream,
                final = final,
                delay = delay,
                force = force,
                callback = callback
            )
        elif self.current == CHUNKED_ENCODING:
            return self.send_chunked(
                data,
                stream = stream,
                final = final,
                delay = delay,
                force = force,
                callback = callback
            )
        elif self.current == GZIP_ENCODING:
            return self.send_gzip(
                data,
                stream = stream,
                final = final,
                delay = delay,
                force = force,
                callback = callback
            )
        elif self.current == DEFLATE_ENCODING:
            return self.send_gzip(
                data,
                stream = stream,
                final = final,
                delay = delay,
                force = force,
                callback = callback
            )

    def send_plain(
        self,
        data,
        stream = None,
        final = True,
        delay = True,
        force = False,
        callback = None
    ):
        return self.send(
            data,
            delay = delay,
            force = force,
            callback = callback
        )

    def send_chunked(
        self,
        data,
        stream = None,
        final = True,
        delay = True,
        force = False,
        callback = None
    ):
        # in case there's no valid data to be sent uses the plain
        # send method to send the empty string and returns immediately
        # to the caller method, to avoid any problems
        if not data: return self.send_plain(
            data,
            stream = stream,
            final = final,
            delay = delay,
            force = force,
            callback = callback
        )

        # creates the new list that is going to be used to store
        # the various parts of the chunk and then calculates the
        # size (in bytes) of the data that is going to be sent
        buffer = []
        size = len(data)

        # creates the various parts of the chunk with the size
        # of the data that is going to be sent and then adds
        # each of the parts to the chunk buffer list
        buffer.append(netius.legacy.bytes("%x\r\n" % size))
        buffer.append(data)
        buffer.append(netius.legacy.bytes("\r\n"))

        # joins the buffer containing the chunk parts and then
        # sends it to the connection using the plain method
        buffer_s = b"".join(buffer)
        return self.send_plain(
            buffer_s,
            stream = stream,
            final = final,
            delay = delay,
            force = force,
            callback = callback
        )

    def send_gzip(
        self,
        data,
        stream = None,
        final = True,
        delay = True,
        force = False,
        callback = None,
        level = 6
    ):
        # verifies if the provided data buffer is valid and in
        # in case it's not propagates the sending to the upper
        # layer (chunked sending) for proper processing
        if not data: return self.send_chunked(
            data,
            stream = stream,
            final = final,
            delay = delay,
            force = force,
            callback = callback
        )

        # "calculates" if the current sending of gzip data is
        # the first one by verifying if the gzip object is set
        is_first = self.gzip == None

        # in case this is the first sending a new compress object
        # is created with the requested compress level, notice that
        # the special deflate case is handled differently
        if is_first:
            is_deflate = self.is_deflate()
            wbits = -zlib.MAX_WBITS if is_deflate else zlib.MAX_WBITS | 16
            self.gzip = zlib.compressobj(level, zlib.DEFLATED, wbits)

        # compresses the provided data string and removes the
        # initial data contents of the compressed data because
        # they are not part of the gzip specification, notice
        # that in case the resulting of the compress operation
        # is not valid a sync flush operation is performed
        data_c = self.gzip.compress(data)
        if not data_c: data_c = self.gzip.flush(Z_PARTIAL_FLUSH)

        # sends the compressed data to the client endpoint setting
        # the correct callback values as requested
        return self.send_chunked(
            data_c,
            stream = stream,
            final = final,
            delay = delay,
            force = force,
            callback = callback
        )

    def set(
        self,
        method,
        url,
        params = None,
        headers = None,
        data = None,
        version = "HTTP/1.1",
        encoding = PLAIN_ENCODING,
        encodings = "gzip, deflate",
        safe = False,
        request = False,
        asynchronous = True,
        timeout = None,
        use_file = False,
        callback = None,
        on_init = None,
        on_open = None,
        on_close = None,
        on_headers = None,
        on_data = None,
        on_result = None,
    ):
        cls = self.__class__

        self.method = method.upper()
        self.url = url
        self.params = params or {}
        self.headers = headers or {}
        self.data = data
        self.version = version
        self.encoding = encoding
        self.current = encoding
        self.encodings = encodings
        self.safe = safe
        self.asynchronous = asynchronous
        self.timeout = timeout or 60
        self.use_file = use_file
        self.parsed = False
        self.request = None
        self.ssl = False
        self.host = None
        self.port = None
        self.path = None
        self.gzip = None
        self.gzip_c = None

        # in case the provided data is a unicode string it's converted into
        # a raw set of bytes using the default encoding
        if netius.legacy.is_unicode(self.data):
            self.data = netius.legacy.bytes(data)

        # in case the data currently set is a plain byte stream wraps it
        # around an iterator to provide compatibility with generators
        if netius.legacy.is_bytes(self.data):
            self.data = iter((len(self.data), self.data))

        # in case the provided data responds to the typical file-like
        # object interface, then it's wrapped around a generator to
        # provided compatibility with the generators based interface
        if hasattr(self.data, "tell"):
            self.data = netius.common.file_iterator(self.data)

        # computes the unique re-usage key for the current protocol taking
        # into account its own state
        self.key = cls.key_g(self.url)

        # runs the unbind all operation to make sure that no handlers is
        # currently registered for operations
        self.unbind_all()

        # in case there's an HTTP parser already set for the protocol runs
        # the reset operation so that its state is guaranteed to be clean
        if self.parser: self.parser.clear()

        # tries to determine if the protocol response should be request
        # wrapped, meaning that a map based object is going to be populated
        # with the contents of the HTTP request/response
        wrap_request = request or (not asynchronous and not on_data and not callback)
        wrap_request = wrap_request or on_result

        # in case the wrap request flag is set (conditions for request usage
        # are met) the protocol is called to run the wrapping operation
        if wrap_request:
            _request, on_close, on_data, callback = self.wrap_request(
                use_file = use_file,
                callback = callback,
                on_close = on_close,
                on_data = on_data,
                on_result = on_result
            )

        # registers for the proper event handlers according to the
        # provided parameters, note that these are considered to be
        # the lower level infra-structure of the event handling
        if on_init: self.bind("loop_set", on_init)
        if on_open: self.bind("open", on_open)
        if on_close: self.bind("close", on_close)
        if on_headers: self.bind("headers", on_headers)
        if on_data: self.bind("partial", on_data)
        if callback: self.bind("message", callback)

        # sets the static part of the protocol internal (no loop is required)
        # so that the required initials fields are properly populated
        self.set_static()

        # returns the current instance with the proper modified state
        # according to the current changes
        return self

    def set_all(self):
        self.set_static()
        self.set_dynamic()

    def set_static(self):
        # creates the message that is going to be used in the logging of
        # the current method request for debugging purposes, this may be
        # useful for a full record traceability of the request
        message = "%s %s %s" % (self.method, self.url, self.version)
        self.debug(message)

        # stores the initial version of the url in a fallback variable so
        # that it may latter be used for the storage of that information
        # in the associated connection (used in callbacks)
        self.base = self.url

        # encodes the provided parameters into the query string and then
        # adds these parameters to the end of the provided url, these
        # values are commonly named get parameters
        query = netius.legacy.urlencode(self.params)
        if query: self.url = self.url + "?" + query

        # parses the provided url and retrieves the various parts of the
        # url that are going to be used in the creation of the connection
        # takes into account some default values in case their are not part
        # of the provided url (eg: port and the scheme)
        self.parsed = netius.legacy.urlparse(self.url)
        self.ssl = self.parsed.scheme == "https"
        self.host = self.parsed.hostname
        self.port = self.parsed.port or (443 if self.ssl else 80)
        self.path = self.parsed.path or "/"
        self.username = self.parsed.username
        self.password = self.parsed.password

        # in case both the username and the password values are defined the
        # authorization header must be created and added to the default set
        # of headers that are going to be included in the request
        if self.username and self.password:
            payload = "%s:%s" % (self.username, self.password)
            payload = netius.legacy.bytes(payload)
            authorization = base64.b64encode(payload)
            authorization = netius.legacy.str(authorization)
            self.headers["authorization"] = "Basic %s" % authorization

        # sets the complete set of information under the protocol instance so
        # that it may be latter used to send the request through the transport
        self.set_headers(self.headers)

    def set_dynamic(self):
        cls = self.__class__

        # creates the function that is going to be used to validate
        # the on connect timeout so that whenever the timeout for
        # the connection operation is exceeded an error is set int
        # the connection and the connection is properly closed
        def connect_timeout():
            if self.is_open(): return
            self.request and cls.set_error(
                "timeout",
                message = "Timeout on connect",
                request = self.request
            )
            self.close()

        # schedules a delay operation to run the timeout handler for
        # both connect operation (this is considered the initial
        # triggers for the such verifiers)
        self.delay(connect_timeout, timeout = self.timeout)

    def run_request(self):
        # retrieves the reference to the top level class to be used
        # for class level operations
        cls = self.__class__

        # saves the current request instance locally to be used latter
        # for request verification (integrity check)
        request = self.request

        # creates a function that is going to be used to validate
        # the receive operation of the connection (receive timeout)
        def receive_timeout():
            # runs the initial verification operations that will
            # try to validate if the requirements for proper request
            # validations are defined, if any of them is not the control
            # full is returned immediately avoiding re-schedule of handler
            if not self.request: return
            if not self.is_open(): return
            if self.request["code"]: return
            if not id(request) == id(self.request): return

            # retrieves the current time and the time of the last data
            # receive operation and using that calculates the delta
            current = time.time()
            last = self.request.get("last", 0)
            delta = current - last

            # retrieves the amount of bytes that have been received so
            # far during the request handling this is going to be used
            # for logging purposes on the error information to be printed
            received = self.request.get("received", 0)

            # determines if the protocol is considered valid, either
            # the connection is not "yet" connected, the time between
            # receive operations is valid or there's data still pending
            # to be sent to the server side, and if that's the case delays
            # the timeout verification according to the timeout value
            if not self.is_open() or delta < self.timeout or\
                not self.transport().get_write_buffer_size() == 0:
                self.delay(receive_timeout, timeout = self.timeout)
                return

            # tries to determine the proper message that is going to be
            # set in the request error, this value should take into account
            # the current development mode flag value
            if self.is_devel(): message = "Timeout on receive (received %d bytes)" % received
            else: message = "Timeout on receive"

            # sets the error information in the request so that the
            # request handler is properly "notified" about the error
            cls.set_error(
                "timeout",
                message = message,
                request = self.request
            )

            # closes the protocol (it's no longer considered valid)
            # and then verifies the various auto closing values
            self.close()

        # sends the request effectively triggering a chain of event
        # that should end with the complete receiving of the response
        self.send_request(callback = lambda c: self.delay(
            receive_timeout, timeout = self.timeout
        ))

    def send_request(self, callback = None):
        method = self.method
        path = self.path
        version = self.version
        headers = self.headers
        data = self.data
        parsed = self.parsed
        safe = self.safe

        if parsed.query: path += "?" + parsed.query

        headers = dict(headers)
        self._apply_base(headers)
        self._apply_dynamic(headers)
        self._apply_connection(headers)
        if safe: self._headers_normalize(headers)

        buffer = []
        buffer.append("%s %s %s\r\n" % (method, path, version))
        for key, value in netius.legacy.iteritems(headers):
            key = netius.common.header_up(key)
            if not isinstance(value, list): value = (value,)
            for _value in value:
                _value = netius.legacy.ascii(_value)
                buffer.append("%s: %s\r\n" % (key, _value))
        buffer.append("\r\n")
        buffer_data = "".join(buffer)

        if data: count = self.send_plain(buffer_data, force = True)
        else: count = self.send_plain(buffer_data, force = True, callback = callback)

        if not data: return count

        def send_part(transport = None):
            try:
                _data = next(data)
            except StopIteration:
                if hasattr(data, "close"): data.close()
                callback and callback(transport)
                return

            self.send_base(_data, force = True, callback = send_part)

        send_part()

        return count

    def wrap_request(
        self,
        use_file = False,
        callback = None,
        on_close = None,
        on_data = None,
        on_result = None
    ):
        """
        Wraps the current set of operations for the protocol so that
        a request object is going to be created and properly populated
        according to the multiple protocol events.

        This method should focus on wrapping the provided callback handlers
        with ones that change the request object state.

        :type use_file: bool
        :param use_file: If a filesystem based approach should be used
        for the storing of the request information.
        :type callback: Function
        :param callback: Callback function to be called when the message
        response has been completely received.
        :type on_close: Function
        :param on_close: Callback function to be called when the underlying
        protocol is closed.
        :type on_data: Function
        :param on_data: Function to be called whenever some data is received
        from the client side, notice that this data may be encoded (eg: gzip).
        :type on_result: Function
        :param on_result: Callback function to be called on the final result
        with the resulting request object.
        :rtype: Tuple
        :return: The tuple containing both the request dictionary object that
        is going to store the information for the request in the current protocol
        and the multiple changed callback methods.
        """

        # retrieves the reference to the parent class object
        # going to be used for class wide operations
        cls = self.__class__

        # saves the references to the previous callback method so that
        # they can be used from the current request based approach
        _on_close = on_close
        _on_data = on_data
        _callback = callback

        # creates both the buffer list and the request structure so that
        # they may be used for the correct construction of the request
        # structure that is going to be send in the callback, then sets
        # the identifier (memory address) of the request in the connection
        buffer = tempfile.NamedTemporaryFile(mode = "w+b") if use_file else []
        self.request = dict(code = None, data = None)

        def on_close(protocol):
            if _on_close: _on_close(protocol)
            protocol._request = None
            if self.request["code"]: return
            cls.set_error(
                "closed",
                message = "Connection closed",
                request = self.request
            )

        def on_data(protocol, parser, data):
            if _on_data: _on_data(protocol, parser, data)
            if use_file: buffer.write(data)
            else: buffer.append(data)
            received = self.request.get("received", 0)
            self.request["received"] = received + len(data)
            self.request["last"] = time.time()

        def callback(protocol, parser, message):
            if _callback: _callback(protocol, parser, message)
            if use_file: cls.set_request_file(parser, buffer, request = self.request)
            else: cls.set_request(parser, buffer, request = self.request)
            if on_result: on_result(protocol, parser, self.request)

        # returns the request object that is going to be properly
        # populated over the life-cycle of the protocol
        return self.request, on_close, on_data, callback

    def set_headers(self, headers, normalize = True):
        self.headers = headers
        if normalize: self.normalize_headers()

    def normalize_headers(self):
        for key, value in netius.legacy.items(self.headers):
            del self.headers[key]
            key = netius.common.header_down(key)
            self.headers[key] = value

    def parse(self, data):
        return self.parser.parse(data)

    def raw_data(self, data):
        """
        Tries to obtain the raw version of the provided data, taking
        into account the possible content encoding present for compression
        or any other kind of operation.

        :type data: String
        :param data: The data to be converted back to its original
        raw value (probably through decompression).
        :rtype: String
        :return: The final raw value for the provided data.
        """

        encoding = self.parser.headers.get("content-encoding", None)
        if not encoding: return data
        if not self.gzip_c:
            is_deflate = encoding == "deflate"
            wbits = zlib.MAX_WBITS if is_deflate else zlib.MAX_WBITS | 16
            self.gzip_c = zlib.decompressobj(wbits)
        return self.gzip_c.decompress(data)

    def is_plain(self):
        return self.current == PLAIN_ENCODING

    def is_chunked(self):
        return self.current > PLAIN_ENCODING

    def is_gzip(self):
        return self.current == GZIP_ENCODING

    def is_deflate(self):
        return self.current == DEFLATE_ENCODING

    def is_compressed(self):
        return self.current > CHUNKED_ENCODING

    def is_uncompressed(self):
        return not self.is_compressed()

    def is_flushed(self):
        return self.current > PLAIN_ENCODING

    def is_measurable(self, strict = True):
        if self.is_compressed(): return False
        if strict and self.is_chunked(): return False
        return True

    def on_data(self, data):
        netius.StreamProtocol.on_data(self, data)
        self.parser.parse(data)

    def _on_data(self):
        message = self.parser.get_message()
        self.trigger("message", self, self.parser, message)
        self.parser.clear()
        self.gzip_c = None

    def on_partial(self, data):
        self.trigger("partial", self, self.parser, data)

    def on_headers(self):
        self.trigger("headers", self, self.parser)

    def on_chunk(self, range):
        self.trigger("chunk", self, self.parser, range)

    def _flush_plain(self, force = False, callback = None):
        if not callback: return
        self.send_plain(b"", force = force, callback = callback)

    def _flush_chunked(self, force = False, callback = None):
        self.send_plain(b"0\r\n\r\n", force = force, callback = callback)

    def _flush_gzip(self, force = False, callback = None):
        # in case the gzip structure has not been initialized
        # (no data sent) no need to run the flushing of the
        # gzip data, so only the chunked part is flushed
        if not self.gzip:
            self._flush_chunked(force = force, callback = callback)
            return

        # flushes the internal zlib buffers to be able to retrieve
        # the data pending to be sent to the client and then sends
        # it using the chunked encoding strategy
        data_c = self.gzip.flush(zlib.Z_FINISH)
        self.send_chunked(data_c, force = force, final = False)

        # resets the gzip values to the original ones so that new
        # requests will starts the information from the beginning
        self.gzip = None

        # runs the flush operation for the underlying chunked encoding
        # layer so that the client is correctly notified about the
        # end of the current request (normal operation)
        self._flush_chunked(force = force, callback = callback)

    def _close_gzip(self, safe = True):
        # in case the gzip object is not defined returns the control
        # to the caller method immediately (nothing to be done)
        if not self.gzip: return

        try:
            # runs the flush operation for the the final finish stage
            # (note that an exception may be raised) and then unsets
            # the gzip object (meaning no more interaction)
            self.gzip.flush(zlib.Z_FINISH)
            self.gzip = None
        except Exception:
            # in case the safe flag is not set re-raises the exception
            # to the caller stack (as expected by the callers)
            if not safe: raise

    def _apply_base(self, headers, replace = False):
        cls = self.__class__
        for key, value in netius.legacy.iteritems(cls.BASE_HEADERS):
            if not replace and key in headers: continue
            headers[key] = value

    def _apply_dynamic(self, headers):
        host = self.host
        port = self.port
        data = self.data
        is_plain = self.is_plain()

        # determines the proper strategy for data payload length, taking into
        # account if there's a payload and if it exists if it's a byte stream
        # or instead an iterator/generator
        if not data: length = 0
        elif netius.legacy.is_bytes(data): length = len(data)
        else: length = next(data)

        # ensures that if the content encoding is plain the content length
        # for the payload is defined otherwise it would be impossible to the
        # server side to determine when the content sending is finished
        netius.verify(
            not is_plain or not length == -1,
            message = "The content length must be defined for plain HTTP encoding"
        )

        if port in (80, 443): host_s = host
        else: host_s = "%s:%d" % (host, port)

        if not "connection" in headers:
            headers["connection"] = "keep-alive"
        if not "host" in headers:
            headers["host"] = host_s
        if not "content-length" in headers and is_plain:
            headers["content-length"] = str(length)
        if not "accept-encoding" in headers and self.encodings:
            headers["accept-encoding"] = self.encodings

    def _apply_connection(self, headers, strict = True):
        is_chunked = self.is_chunked()
        is_gzip = self.is_gzip()
        is_deflate = self.is_deflate()
        is_compressed = self.is_compressed()
        is_measurable = self.is_measurable(strict = strict)
        has_length = "content-length" in headers
        has_ranges = "accept-ranges" in headers

        if is_chunked: headers["transfer-encoding"] = "chunked"
        if is_gzip: headers["content-encoding"] = "gzip"
        if is_deflate: headers["content-encoding"] = "deflate"

        if not is_measurable and has_length: del headers["content-length"]
        if is_compressed and has_ranges: del headers["accept-ranges"]

    def _headers_normalize(self, headers):
        for key, value in netius.legacy.items(headers):
            if not type(value) in (list, tuple): continue
            headers[key] = ";".join(value)

class HTTPClient(netius.ClientAgent):
    """
    Simple test of an HTTP client, supports a series of basic
    operations and makes use of the HTTP parser from netius.

    The current implementation supports the auto-release of the
    connection once the message has been received, this is optional
    and may be disabled with an argument in the constructor.
    """

    protocol = HTTPProtocol

    def __init__(
        self,
        auto_release = True,
        *args,
        **kwargs
    ):
        netius.ClientAgent.__init__(self, *args, **kwargs)
        self.auto_release = auto_release
        self.available = dict()
        self._loop = None

    @classmethod
    def get_s(
        cls,
        url,
        params = {},
        headers = {},
        **kwargs
    ):
        return cls.method_s(
            "GET",
            url,
            params = params,
            headers = headers,
            **kwargs
        )

    @classmethod
    def post_s(
        cls,
        url,
        params = {},
        headers = {},
        data = None,
        **kwargs
    ):
        return cls.method_s(
            "POST",
            url,
            params = params,
            headers = headers,
            data = data,
            **kwargs
        )

    @classmethod
    def put_s(
        cls,
        url,
        params = {},
        headers = {},
        data = None,
        **kwargs
    ):
        return cls.method_s(
            "PUT",
            url,
            params = params,
            headers = headers,
            data = data,
            **kwargs
        )

    @classmethod
    def delete_s(
        cls,
        url,
        params = {},
        headers = {},
        **kwargs
    ):
        return cls.method_s(
            "DELETE",
            url,
            params = params,
            headers = headers,
            **kwargs
        )

    @classmethod
    def method_s(
        cls,
        method,
        url,
        params = {},
        headers = {},
        data = None,
        version = "HTTP/1.1",
        safe = False,
        asynchronous = True,
        daemon = True,
        timeout = None,
        ssl_verify = False,
        use_file = False,
        callback = None,
        on_init = None,
        on_open = None,
        on_close = None,
        on_headers = None,
        on_data = None,
        on_result = None,
        http_client = None,
        **kwargs
    ):
        # in case no HTTP client instance is provided tries to
        # retrieve a static global one (singleton) to be used
        # for the current request operation
        if not http_client:
            http_client = cls.get_client_s(
                daemon = daemon,
                **kwargs
            )

        # calls the underlying method on the current HTTP client
        # propagating most of the arguments, and retrieves the resulting
        # value to be propagated to the current method's caller
        result = http_client.method(
            method,
            url,
            params = params,
            headers = headers,
            data = data,
            version = version,
            safe = safe,
            asynchronous = asynchronous,
            timeout = timeout,
            ssl_verify = ssl_verify,
            use_file = use_file,
            callback = callback,
            on_init = on_init,
            on_open = on_open,
            on_close = on_close,
            on_headers = on_headers,
            on_data = on_data,
            on_result = on_result,
            **kwargs
        )

        # returns the "final" result to the caller method so that
        # it may be used/processed by it (as expected)
        return result

    @classmethod
    def to_response(cls, map, raise_e = True):
        """
        Simple utility method that takes the classic dictionary
        based request and converts it into a simple HTTP response
        object to be used in a more interactive way.

        :type map: Dictionary
        :param map: The dictionary backed request object that is
        going to be converted into a response.
        :type raise_e: bool
        :param raise_e: If an exception should be raised in case
        there's an error in the HTTP status field.
        :rtype: HTTPResponse
        :return: The normalized response value.
        """

        error = map.get("error", None)
        message = map.get("message", None)
        exception = map.get("exception", None)
        is_error = True if error and raise_e else False
        if not is_error: return netius.common.HTTPResponse(
            data = map.get("data", None),
            code = map.get("code", 500),
            status = map.get("status", None),
            headers = map.get("headers", None)
        )
        message = message or "Undefined error (%s)" % error
        if exception: raise exception
        raise netius.NetiusError(message)

    def cleanup(self):
        netius.ClientAgent.cleanup(self)

        # iterates over the complete set of protocol instances
        # to be re-used and closes them, then empties the map
        # of available instances (no more re-usage possible)
        for protocol in netius.legacy.values(self.available):
            protocol.close()
        self.available.clear()

        # in case a (static) event loop is defined closes it,
        # not allowing any further re-usage of it (as expected)
        self._close_loop()

    def get(
        self,
        url,
        params = {},
        headers = {},
        **kwargs
    ):
        return self.method(
            "GET",
            url,
            params = params,
            headers = headers,
            **kwargs
        )

    def post(
        self,
        url,
        params = {},
        headers = {},
        data = None,
        **kwargs
    ):
        return self.method(
            "POST",
            url,
            params = params,
            headers = headers,
            data = data,
            **kwargs
        )

    def put(
        self,
        url,
        params = {},
        headers = {},
        data = None,
        **kwargs
    ):
        return self.method(
            "PUT",
            url,
            params = params,
            headers = headers,
            data = data,
            **kwargs
        )

    def delete(
        self,
        url,
        params = {},
        headers = {},
        **kwargs
    ):
        return self.method(
            "DELETE",
            url,
            params = params,
            headers = headers,
            **kwargs
        )

    def method(
        self,
        method,
        url,
        params = None,
        headers = None,
        data = None,
        version = "HTTP/1.1",
        encoding = PLAIN_ENCODING,
        encodings = "gzip, deflate",
        safe = False,
        request = False,
        close = True,
        asynchronous = True,
        timeout = None,
        ssl_verify = False,
        use_file = False,
        callback = None,
        on_init = None,
        on_open = None,
        on_close = None,
        on_headers = None,
        on_data = None,
        on_result = None,
        loop = None,
        **kwargs
    ):
        # extracts the reference to the upper class element associated
        # with the current instance, to be used for operations
        cls = self.__class__

        # tries to retrieve the unique key from the provided URL and then
        # uses it to try to retrieve a possible already available protocol,
        # for connection re-usage (avoids long establish connection times)
        # notice that the event loop is also re-used accordingly
        key = cls.protocol.key_g(url)
        protocol = self.available.pop(key, None)
        if protocol and (not protocol.is_open() or\
            protocol.transport().is_closing()): protocol = None
        if protocol: loop = loop or protocol.loop()

        # determines if the loop instance was provided by the user so
        # that latter on we can determine if it should be closed (garbage
        # collection or not)
        user_loop = True if loop else False

        # in case the current execution model is not asynchronous a new
        # loop context must exist otherwise it may collide with the global
        # event loop execution creating unwanted behaviour
        if not asynchronous: loop = loop or self._get_loop(**kwargs)

        # creates the new protocol instance that is going to be used to
        # handle this new request, a new protocol represents also a new
        # parser object as defined by structure
        callable = protocol.set if protocol else cls.protocol
        protocol = callable(
            method,
            url,
            params = params,
            headers = headers,
            data = data,
            version = version,
            encoding = encoding,
            encodings = encodings,
            safe = safe,
            request = request,
            asynchronous = asynchronous,
            timeout = timeout,
            use_file = use_file,
            callback = callback,
            on_init = on_init,
            on_open = on_open,
            on_close = on_close,
            on_headers = on_headers,
            on_data = on_data,
            on_result = on_result
        )

        # verifies if the current protocol is already open and if that's the
        # case calls the connection made directly, indicating that the connection
        # is already established (re-usage of protocol)
        if protocol.is_open():
            protocol.connection_made(protocol.transport())

        # runs the global connect stream function on netius to initialize the
        # connection operation and maybe a new event loop (if that's required)
        else:
            loop = netius.connect_stream(
                lambda: protocol,
                protocol.host,
                protocol.port,
                ssl = protocol.ssl,
                ssl_verify = ssl_verify,
                loop = loop
            )

        # in case the asynchronous mode is enabled returns the loop and the protocol
        # immediately so that it can be properly used by the caller
        if asynchronous: return loop, protocol

        def on_message(protocol, parser, message):
            # in case the auto release (no connection re-usage) mode is
            # set the protocol is closed immediately
            if self.auto_release:
                protocol.close()

            # verifies if the current connection is meant to be kept alive
            # and if that's not the case closes it immediately, this way
            # the client is the responsible for triggering the disconnect
            # operation, avoiding problems with possible connection re-usage
            elif not parser.keep_alive:
                protocol.close()

            # otherwise the protocol is set in the available map and
            # the only the loop is stopped (unblocking the processor)
            else:
                self.available[protocol.key] = protocol
                netius.compat_loop(loop).stop()

        def on_close(protocol):
            # verifies if the protocol being closed is currently in
            # the pool of available protocols, so that decisions on
            # the stopping of the event loop may be made latter on
            from_pool = protocol.key in self.available

            # because the protocol was closed we must release it from
            # the available map (if it exits) and then unblock the current
            # event loop call (stop operation)
            self.available.pop(protocol.key, None)

            # in case the protocol that is being closed is not the one
            # in usage returns immediately (no need to stop the event
            # loop for a protocol from the available pool)
            if from_pool: return

            # tries to retrieve the loop compatible value and if it's
            # successful runs the stop operation on the loop
            netius.compat_loop(loop).stop()

        # binds the protocol message and finish events to the associated
        # function for proper handling of the synchronous request details
        protocol.bind("message", on_message)
        protocol.bind("close", on_close)

        # runs the loop until complete, this should be the main blocking
        # call into the event loop, notice that in case the loop that was
        # used is not the HTTP client's static loop and also not a user's
        # provided loop it's closed immediately (garbage collection)
        loop.run_forever()
        if not loop == self._loop and not user_loop: loop.close()

        # returns the final request object (that should be populated by this
        # time) to the called method, so that a simple interface is provided
        return protocol.request

    def _get_loop(self, **kwargs):
        if not self._loop: self._loop = netius.new_loop(**kwargs)
        return self._loop

    def _close_loop(self):
        if not self._loop: return
        self._loop.close()
        self._loop = None

if __name__ == "__main__":
    buffer = []

    def on_headers(protocol, parser):
        print(parser.code_s + " " + parser.status_s)

    def on_partial(protocol, parser, data):
        data = data
        data and buffer.append(data)

    def on_message(protocol, parser, message):
        request = HTTPProtocol.set_request(parser, buffer)
        print(request["headers"])
        print(request["data"] or b"[empty data]")
        protocol.close()

    def on_finish(protocol):
        netius.compat_loop(loop).stop()

    url = netius.conf("HTTP_URL", "https://www.flickr.com/")

    client = HTTPClient()
    loop, protocol = client.get(url)

    protocol.bind("headers", on_headers)
    protocol.bind("partial", on_partial)
    protocol.bind("message", on_message)
    protocol.bind("finish", on_finish)

    loop.run_forever()
    loop.close()
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import tempfile

import netius

from . import util
from . import parser

FILE_LIMIT = 5242880
""" The limit value (in bytes) from which the back-end
message storage mechanism will start using a file system
stored file instead of an in memory object, this way it's
possible to avoid memory starvation problems, this is a
default value for the parser and may be overriden using
the dedicated parameter value in the constructor """

REQUEST = 1
""" The HTTP request message indicator, should be
used when identifying the HTTP request messages """

RESPONSE = 2
""" Indicator for the response HTTP message, can be
used to identify chunks of data that represent an
HTTP based response """

LINE_STATE = 1
""" The initial state where the status line is meant
to be read as expected by the HTTP specification """

HEADERS_STATE = 2
""" State here the headers are waiting to be read this
state exits when the end of headers sequence is found
and the headers are parsed """

MESSAGE_STATE = 3
""" The message state for the situations where the message
is waiting to be processed (read) this state may persist
for some time if the message is big enough """

FINISH_STATE = 4
""" The final state set when the complete HTTP request or
response has been processed, if a parse operation starts
with this state the parsed is reseted """

PLAIN_ENCODING = 1
""" Plain text encoding that does not transform the
data from its based format, should be used only as
a fallback method because of performance issues """

CHUNKED_ENCODING = 2
""" Chunked based encoding that allows the sending of
the data as a series of length based parts """

GZIP_ENCODING = 3
""" The gzip based encoding used to compress data, this
kind of encoding will always used chunked encoding so
that the content may be send in parts """

DEFLATE_ENCODING = 4
""" The deflate based encoding used to compress data, this
kind of encoding will always used chunked encoding so
that the content may be send in parts """

HTTP_09 = 1
""" The enumeration value for the temporary and beta HTTP
version 0.9 version (very rare) """

HTTP_10 = 2
""" Value for the first version of the HTTP specification,
connection running under this version should be closed right
away as defined in the specification """

HTTP_11 = 3
""" Current version of the HTTP specification should be the
most commonly used nowadays, connection running under this
version of the protocol should keep connections open """

VERSIONS_MAP = {
    "HTTP/0.9" : HTTP_09,
    "HTTP/1.0" : HTTP_10,
    "HTTP/1.1" : HTTP_11
}
""" Maps associating the standard HTTP version string with the
corresponding enumeration based values for each of them """

CODE_STRINGS = {
    100 : "Continue",
    101 : "Switching Protocols",
    200 : "OK",
    201 : "Created",
    202 : "Accepted",
    203 : "Non-Authoritative Information",
    204 : "No Content",
    205 : "Reset Content",
    206 : "Partial Content",
    207 : "Multi-Status",
    301 : "Moved permanently",
    302 : "Found",
    303 : "See Other",
    304 : "Not Modified",
    305 : "Use Proxy",
    306 : "(Unused)",
    307 : "Temporary Redirect",
    400 : "Bad Request",
    401 : "Unauthorized",
    402 : "Payment Required",
    403 : "Forbidden",
    404 : "Not Found",
    405 : "Method Not Allowed",
    406 : "Not Acceptable",
    407 : "Proxy Authentication Required",
    408 : "Request Timeout",
    409 : "Conflict",
    410 : "Gone",
    411 : "Length Required",
    412 : "Precondition Failed",
    413 : "Request Entity Too Large",
    414 : "Request-URI Too Long",
    415 : "Unsupported Media Type",
    416 : "Requested Range Not Satisfiable",
    417 : "Expectation Failed",
    500 : "Internal Server Error",
    501 : "Not Implemented",
    502 : "Bad Gateway",
    503 : "Service Unavailable",
    504 : "Gateway Timeout",
    505 : "HTTP Version Not Supported"
}
""" Dictionary associating the error code as integers
with the official descriptive message for it """

class HTTPParser(parser.Parser):
    """
    Parser object for the HTTP format, should be able to
    parse both request and response messages.

    The parser itself should be event driven an callback
    functions should be called upon partials information
    parsing. But the object itself is not thread safe.
    """

    FIELDS = (
        "_pid",
        "type",
        "store",
        "file_limit",
        "state",
        "buffer",
        "headers",
        "message",
        "method",
        "version",
        "code",
        "keep_alive",
        "line_s",
        "headers_s",
        "method_s",
        "path_s",
        "version_s",
        "code_s",
        "status_s",
        "connection_s",
        "message_s",
        "message_f",
        "content_l",
        "message_l",
        "transfer_e",
        "encodings",
        "chunked",
        "chunk_d",
        "chunk_l",
        "chunk_s",
        "chunk_e"
    )

    def __init__(
        self,
        owner,
        type = REQUEST,
        store = False,
        file_limit = FILE_LIMIT
    ):
        parser.Parser.__init__(self, owner)

        self.build()
        self.reset(type = type, store = store, file_limit = file_limit)

    def build(self):
        """
        Builds the initial set of states ordered according to
        their internal integer definitions, this method provides
        a fast and scalable way of parsing data.
        """

        parser.Parser.build(self)

        self.connection = self.owner

        self.states = (
            self._parse_line,
            self._parse_headers,
            self._parse_message
        )
        self.state_l = len(self.states)

    def destroy(self):
        """
        Destroys the current structure for the parser meaning that
        it's restored to the original values, this method should only
        be called on situation where no more parser usage is required.
        """

        parser.Parser.destroy(self)

        self.clear()
        self.close()

        self.connection = None
        self.states = ()
        self.state_l = 0

    def reset(self, type = REQUEST, store = False, file_limit = FILE_LIMIT):
        """
        Initializes the state of the parser setting the values
        for the various internal structures to the original value.
        After this operation the parser is ready for a new parse.

        :type type: int
        :param type: The type of HTTP message that is going to be
        parsed using the current parser.
        :type store: bool
        :param store: If the complete message body should be stored
        in memory as the message gets loaded, this option may create
        some serious performance issues.
        :type file_limit: int
        :param file_limit: The maximum content for the payload message
        from which a in file buffer will be used instead of the one that
        is stored in memory (avoid memory starvation).
        """

        self.close()
        self.type = type
        self.store = store
        self.file_limit = file_limit
        self.state = LINE_STATE
        self.buffer = []
        self.headers = {}
        self.message = []
        self.method = None
        self.version = None
        self.code = None
        self.keep_alive = False
        self.line_s = None
        self.headers_s = None
        self.method_s = None
        self.path_s = None
        self.version_s = None
        self.code_s = None
        self.status_s = None
        self.connection_s = None
        self.message_s = None
        self.message_f = None
        self.content_l = -1
        self.message_l = 0
        self.transfer_e = None
        self.encodings = None
        self.chunked = False
        self.chunk_d = 0
        self.chunk_l = 0
        self.chunk_s = 0
        self.chunk_e = 0

    def clear(self, force = False):
        if not force and self.state == LINE_STATE: return
        self.reset(
            type = self.type,
            store = self.store,
            file_limit = self.file_limit
        )

    def close(self):
        if hasattr(self, "message") and self.message: self.message = []
        if hasattr(self, "message_f") and self.message_f: self.message_f.close()

    def get_path(self, normalize = False):
        """
        Retrieves the path associated with the request, this
        value should be interpreted from the HTTP status line.

        In case the normalize flag is set a possible absolute
        URL value should be normalized into an absolute path.
        This may be required under some proxy related scenarios.

        :type normalize: bool
        :param normalize: If the normalization process should be
        applied for absolute URL scenarios.
        :rtype: String
        :return: The path associated with the current request.
        """

        split = self.path_s.split("?", 1)
        path = split[0]
        if not normalize: return path
        if not path.startswith(("http://", "https://")): return path
        return netius.legacy.urlparse(path).path

    def get_query(self):
        """
        Retrieves the (GET) query part of the path, this is considered
        to be the part of the path after the first question mark.

        This query string may be used to parse any possible (GET)
        arguments.

        :rtype: String
        :return: The query part of the path, to be used for parsing
        of (GET) arguments.
        """

        split = self.path_s.split("?", 1)
        if len(split) == 1: return ""
        else: return split[1]

    def get_message(self):
        """
        Gathers the complete set of message contents for the current
        request/response in parsing. The proper gathering strategy will
        depend on the current strategy (eg: in memory vs file strategies).

        The result of this process is cached meaning that further calls
        to this method will return the same result.

        This method should be used carefully as it may create some memory
        performance issues when retrieving large message values.

        :rtype: String
        :return: The message for the current parsing process as a linear
        string value that may be used as a simple buffer.
        """

        if self.message_s: return self.message_s
        if self.message_f: self.message_s = self.get_message_f()
        else: self.message_s = b"".join(self.message)
        return self.message_s

    def get_message_f(self):
        self.message_f.seek(0)
        return self.message_f.read()

    def get_message_b(self, copy = False, size = 40960):
        """
        Retrieves a new buffer associated with the currently
        loaded message, the first time this method is called a
        new in memory object will be created for the storage.

        In case the current parsing operation is using a file like
        object for the handling this object it is returned instead.

        The call of this method is only considered to be safe after
        the complete message has been received and processed, otherwise
        and invalid message file structure may be created.

        Note that the returned object will always be set at the
        beginning of the file, so some care should be taken in usage.

        :type copy: bool
        :param copy: If a copy of the file object should be returned
        or if instead the shallow copy associated with the parser should
        be returned instead, this should be used carefully to avoid any
        memory leak from file descriptors.
        :type size: int
        :param size: Size (in bytes) of the buffer to be used in a possible
        copy operation between buffers.
        :rtype: File
        :return: The file like object that may be used to percolate
        over the various parts of the current message contents.
        """

        # in case there's not message file currently enabled creates one
        # and writes the value of the message into it
        if not self.message_f:
            self.message_f = netius.legacy.BytesIO()
            for value in self.message: self.message_f.write(value)

        # restores the message file to the original/initial position and
        # then in case there's no copy required returns it immediately
        self.message_f.seek(0)
        if not copy: return self.message_f

        # determines if the file limit for a temporary file has been
        # surpassed and if that's the case creates a named temporary
        # file, otherwise created a memory based buffer
        use_file = self.store and self.content_l >= self.file_limit
        if use_file: message_f = tempfile.NamedTemporaryFile(mode = "w+b")
        else: message_f = netius.legacy.BytesIO()

        try:
            # iterates continuously reading the contents from the message
            # file and writing them back to the output (copy) file
            while True:
                data = self.message_f.read(size)
                if not data: break
                message_f.write(data)
        finally:
            # resets both of the message file (output and input) to the
            # original position as expected by the infra-structure
            self.message_f.seek(0)
            message_f.seek(0)

        # returns the final (copy) of the message file to the caller method
        # note that the type of this file may be an in memory or stored value
        return message_f

    def get_headers(self):
        headers = dict(self.headers)
        for key, value in netius.legacy.iteritems(self.headers):
            key_up = util.header_up(key)
            del headers[key]
            headers[key_up] = value
        return headers

    def get_encodings(self):
        if not self.encodings == None: return self.encodings
        accept_encoding_s = self.headers.get("accept-encoding", "")
        self.encodings = [value.strip() for value in accept_encoding_s.split(",")]
        return self.encodings

    def parse(self, data):
        """
        Parses the provided data chunk, changing the current
        state of the parser accordingly and returning the
        number of processed bytes from it.

        :type data: String
        :param data: The string containing the data to be parsed
        in the current parse operation.
        :rtype: int
        :return: The amount of bytes of the data string that have
        been "parsed" in the current parse operation.
        """

        parser.Parser.parse(self, data)

        # in case the current state of the parser is finished, must
        # reset the state to the start position as the parser is
        # re-starting (probably a new data sequence)
        if self.state == FINISH_STATE: self.clear()

        # retrieves the size of the data that has been sent for parsing
        # and saves it under the size original variable
        size = len(data)
        size_o = size

        # iterates continuously to try to process all that
        # data that has been sent for processing
        while size > 0:

            # iterates while the current state is valid for
            # parsing as there are only parsing methods for
            # the range of valid states
            if self.state <= self.state_l:

                # retrieves the parsing method for the current
                # state and then runs it retrieving the number
                # of valid parsed bytes in case this value is
                # zero the parsing iteration is broken
                method = self.states[self.state - 1]
                count = method(data)
                if count == 0: break

                # decrements the size of the data buffer by the
                # size of the parsed bytes and then retrieves the
                # sub part of the data buffer as the new data buffer
                size -= count
                data = data[count:]

                # continues the loop as there should be still some
                # data remaining to be parsed in the current buffer
                continue

            elif self.state == FINISH_STATE:

                # must clear the current parser state, so that it may
                # start the parsing of a new message and then continue
                # the loop trying to find new contents for parsing, this
                # critical for HTTP pipelining support
                self.clear()
                continue

            else:
                raise netius.ParserError("Invalid state '%d'" % self.state)

        # in case not all of the data has been processed
        # must add it to the buffer so that it may be used
        # latter in the next parsing of the message
        if size > 0: self.buffer.append(data)

        # returns the number of read (processed) bytes of the
        # data that has been sent to the parser
        return size_o - size

    def _parse_line(self, data):
        # tries to find the final newline value in the provided
        # data in case there's one it's considered that the the
        # initial line must have been found
        index = data.find(b"\n")
        if index == -1: return 0

        # adds the partial data (until line ending) to the buffer
        # and then joins the buffer as the initial line, this value
        # should not include the final newline characters, after that
        # the buffer is cleared as new data is going to be stored for
        # (remaining part of the request or response)
        self.buffer.append(data[:index])
        self.line_s = b"".join(self.buffer).rstrip()
        self.line_s = netius.legacy.str(self.line_s)
        del self.buffer[:]

        # restores the final end of line sequence to the buffer, this
        # allows "simple requests" to be parsed properly in under the
        # next section of parsing headers (required for compliance)
        self.buffer.append(b"\r\n")

        # splits the line around its various components, verifying than
        # that the number of provided items is the expected one, notice
        # that for responses the parsing is relaxed as the status string
        # can be an empty string (no message to be presented)
        values = self.line_s.split(" ", 2)
        if self.type == RESPONSE and len(values) == 2: values.append("")
        if not len(values) == 3: raise netius.ParserError(
            "Invalid status line '%s'" % self.line_s
        )

        # determines if the current type of parsing is request based
        # and if that's the case unpacks the status line as a request
        if self.type == REQUEST:
            self.method_s, self.path_s, self.version_s = values
            self.method = self.method_s.lower()
            self.version = VERSIONS_MAP.get(self.version_s, HTTP_10)

        # otherwise ensures that the parsing type is response based
        # and unpacks the status line accordingly
        elif self.type == RESPONSE:
            self.version_s, self.code_s, self.status_s = values
            self.version = VERSIONS_MAP.get(self.version_s, HTTP_10)
            self.code = int(self.code_s)
            self.status = self.status_s

        # updates the current state of parsing to the header state
        # so that the headers are the next item to be processed
        self.state = HEADERS_STATE

        # triggers the on line event so that the listeners are notified
        # about the end of the parsing of the status line and then
        # returns the count of the parsed bytes of the message
        self.trigger("on_line")
        return index + 1

    def _parse_headers(self, data):
        # creates a temporary buffer with the contents of
        # the current buffer plus the current data and then
        # joins it to retrieve the current complete buffer
        # string to be used in the finding of the end of
        # header sequence (required for complete parsing)
        buffer_t = self.buffer + [data]
        buffer_s = b"".join(buffer_t)

        # tries to find the end of headers sequence in case
        # it's not found returns the zero value meaning that
        # the no bytes have been processed (delays parsing)
        index = buffer_s.find(b"\r\n\r\n")
        if index == -1: return 0

        # retrieves the partial headers string from the buffer
        # string and then deletes the current buffer so that
        # it may be reused for other partial parsings
        self.headers_s = buffer_s[:index]
        del self.buffer[:]

        # calculates the base length as the difference between
        # the buffer string and the length of the currently
        # provided data value and then uses this value to calculate
        # the base index for the process data count value
        base_length = len(buffer_s) - len(data)
        base_index = index - base_length

        # splits the complete set of lines that compromise
        # the headers and then iterates over each of them
        # to set the key and value in the headers map
        lines = self.headers_s.split(b"\r\n")
        for line in lines:
            # verifies if the line contains any information if
            # that's not the case the current cycle must be
            # skipped as this may be an extra empty line
            if not line: continue

            # tries to split the line around the key to value
            # separator in case there's no valid split (two
            # values were not found) an exception must be raised
            values = line.split(b":", 1)
            if not len(values) == 2:
                raise netius.ParserError("Invalid header line")

            # unpacks the values an normalizes them, lowering the
            # case of the key and stripping the values of any extra
            # whitespace like value that may exist in them
            key, value = values
            key = key.strip().lower()
            key = netius.legacy.str(key)
            value = value.strip()
            value = netius.legacy.str(value, errors = "replace")
            exists = key in self.headers

            # in case the header already exists this indicates that
            # there are multiple definitions of the header and a sequence
            # must be used in order to store the various headers
            if exists:
                sequence = self.headers[key]
                is_list = type(sequence) == list
                if not is_list: sequence = [sequence]
                sequence.append(value)
                value = sequence

            # sets the final header value into the headers map so that
            # it may be used latter for the serialization process
            self.headers[key] = value

        # retrieves the size of the contents from the populated
        # headers, this is not required by the specification and
        # the parser should be usable even without it
        self.content_l = self.headers.get("content-length", -1)
        self.content_l = self.content_l and int(self.content_l)

        # verifies if a back-end file object should be used to store
        # the file contents, this is done by checking the store flag
        # and verifying that the file limit value has been reached
        use_file = self.store and self.content_l >= self.file_limit
        if use_file: self.message_f = tempfile.NamedTemporaryFile(mode = "w+b")

        # retrieves the type of transfer encoding that is going to be
        # used in the processing of this request in case it's of type
        # chunked sets the current chunked flag indicating that the
        # request is meant to be processed as so
        self.transfer_e = self.headers.get("transfer-encoding", None)
        self.chunked = self.transfer_e == "chunked"

        # in case the current response in parsing has the no content
        # code (no payload present) the content length is set to the
        # zero value in case it has not already been populated
        if self.type == RESPONSE and self.code in (204, 304) and\
            self.content_l == -1: self.content_l = 0

        # in case the current request is not chunked and the content length
        # header is not defined the content length is set to zero because
        # for normal requests with payload the content length is required
        # and if it's omitted it means there's no payload present
        if self.type == REQUEST and not self.chunked and\
            self.content_l == -1: self.content_l = 0

        # verifies if the connection is meant to be kept alive by
        # verifying the current value of the connection header against
        # the expected keep alive string value, note that the verification
        # takes into account a possible list value in connection
        self.connection_s = self.headers.get("connection", None)
        if type(self.connection_s) == list: self.connection_s = self.connection_s[0]
        self.connection_s = self.connection_s and self.connection_s.lower()
        self.keep_alive = self.connection_s == "keep-alive"
        self.keep_alive |= self.connection_s == None and self.version >= HTTP_11

        # verifies if the current message has finished, for those
        # situations an extra (finish) state change will be issued
        has_finished = self.content_l == 0

        # updates the current state of parsing to the message state
        # as that the headers are followed by the message
        if has_finished: self.state = FINISH_STATE
        else: self.state = MESSAGE_STATE

        # triggers the on headers event so that the listener object
        # is notified about the parsing of the headers and than returns
        # the parsed amount of information (bytes) to the caller
        self.trigger("on_headers")
        if has_finished: self.trigger("on_data")
        return base_index + 4

    def _parse_message(self, data):
        if self.chunked: return self._parse_chunked(data)
        else: return self._parse_normal(data)

    def _parse_normal(self, data):
        # retrieves the size of the data that has just been
        # received and then in case the store flag is set
        # stores the data in the proper buffer and increments
        # the message length counter with the size of the data
        data_l = len(data)
        if self.store: self._store_data(data)
        self.message_l += data_l

        # verifies if the complete message has already been
        # received, that occurs if the content length is
        # defined and the value is the is the same as the
        # currently defined message length
        has_finished = not self.content_l == -1 and\
            self.message_l == self.content_l

        # triggers the partial data received event and then
        # in case the complete message has not been received
        # returns immediately the length of processed data
        self.trigger("on_partial", data)
        if not has_finished: return data_l

        # updates the current state to the finish state and then
        # triggers the on data event (indicating the end of the
        # parsing of the message)
        self.state = FINISH_STATE
        self.trigger("on_data")

        # returns the length of the processed data as the amount
        # of processed bytes by the current method
        return data_l

    def _parse_chunked(self, data):
        # starts the parsed byte counter with the initial zero
        # value this will be increment as bytes are parsed
        count = 0

        # verifies if the end of chunk state has been reached
        # that happens when only the last two characters remain
        # to be parsed from the chunk
        is_end = self.chunk_l < 3 and self.chunk_l > 0
        if is_end:
            # calculates the size of the data that is going
            # to be parsed as that's required to check if
            # the end chunk state has been reached
            data_l = len(data)

            # in case the required amount of data has not
            # been received returns the parsed bytes amount
            # (count) immediately to the caller
            if data_l < self.chunk_l:
                self.chunk_l -= data_l
                count += data_l
                return count

            # adds the parsed number of bytes to the count value,
            # resets the pending length of the chunk to the initial
            # zero value and then returns that value to the caller
            count += self.chunk_l
            self.chunk_l = 0

            # in case the current chunk dimension (size) is
            # zero this is the last chunk and so the state
            # must be set to the finish and the on data event
            # must be triggered to indicate the end of message
            if self.chunk_d == 0:
                self.state = FINISH_STATE
                self.trigger("on_data")

            # otherwise this is the end of a "normal" chunk and
            # and so the end of chunk index must be calculated
            # and the chunk event must be triggered
            else:
                self.chunk_e = len(self.message)
                self.trigger("on_chunk", (self.chunk_s, self.chunk_e))

            # in case the message is not meant to be stored or in
            # case the file storage mode is active (spares memory),
            # deletes the contents of the message buffer as they're
            # not going to be used to access request's data as a whole
            if not self.store or self.message_f: del self.message[:]

            # returns the number of bytes that have been parsed by
            # the current end of chunk operation to the caller method
            return count

        # check if the start of the chunk state is the current one
        # does that by verifying that the current value for the chunk
        # length is zero (initial situation)
        is_start = self.chunk_l == 0
        if is_start:
            # tries to find the separator of the initial value for
            # the chunk in case it's not found returns immediately
            index = data.find(b"\n")
            if index == -1: return 0

            # some of the current data to the buffer and then re-joins
            # it as the header value, then removes the complete set of
            # contents from the buffer so that it may be re-used
            self.buffer.append(data[:index])
            header = b"".join(self.buffer)[:-1]
            del self.buffer[:]

            # sets the new data buffer as the partial buffer of the data
            # except the extra newline character (not required)
            data = data[index + 1:]

            # splits the header value so that additional chunk information
            # is removed and then parsed the value as the original chunk
            # size (dimension) adding the two extra bytes to the length
            header_s = header.split(b";", 1)
            size = header_s[0]
            self.chunk_d = int(size.strip(), base = 16)
            self.chunk_l = self.chunk_d + 2
            self.chunk_s = len(self.message)

            # increments the counter of the parsed number of bytes from the
            # provided data by the index of the newline character position
            # plus one byte respecting to the newline character
            count += index + 1

        # retrieves the partial data that is valid according to the
        # calculated chunk length and then calculates the size of
        # "that" partial data string value
        data = data[:self.chunk_l - 2]
        data_s = len(data)

        # adds the partial data to the message list and runs the store operation
        # just in case the storage of the data in file is required, then decrements
        # the (remaining) chunk length by the size of the read data, note that
        # the message buffer is used even if the store flag is not set, so that
        # it's possible to refer the chunk as a tuple of start and end indexes when
        # triggering the chunk parsed (on chunk) event (performance gains)
        if data: self.message.append(data)
        if data and self.store: self._store_data(data, memory = False)
        self.chunk_l -= data_s

        # in case there's data parsed the partial data event
        # is triggered to notify handlers about the new data
        if data: self.trigger("on_partial", data)

        # increments the byte counter value by the size of the data
        # and then returns the same counter to the caller method
        count += data_s
        return count

    def _parse_query(self, query):
        # runs the "default" parsing of the query string from the system
        # and then decodes the complete set of parameters properly
        params = netius.legacy.parse_qs(query, keep_blank_values = True)
        return self._decode_params(params)

    def _store_data(self, data, memory = True):
        if not self.store: raise netius.ParserError("Store is not possible")
        if self.message_f: self.message_f.write(data)
        elif memory: self.message.append(data)

    def _decode_params(self, params):
        _params = dict()

        for key, value in netius.legacy.iteritems(params):
            items = []
            for item in value:
                is_bytes = netius.legacy.is_bytes(item)
                if is_bytes: item = item.decode("utf-8")
                items.append(item)
            is_bytes = netius.legacy.is_bytes(key)
            if is_bytes: key = key.decode("utf-8")
            _params[key] = items

        return _params

class HTTPResponse(object):

    def __init__(self, data = None, code = 200, status = None, headers = None):
        self.data = data
        self.code = code
        self.status = status
        self.headers = headers

    def read(self):
        return self.data

    def readline(self):
        return self.read()

    def close(self):
        pass

    def getcode(self):
        return self.code

    def info(self):
        return self.headers



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os

from . import base

class PasswdAuth(base.Auth):

    def __init__(self, path = None, *args, **kwargs):
        base.Auth.__init__(self, *args, **kwargs)
        self.path = path

    @classmethod
    def auth(cls, username, password, path = "passwd", *args, **kwargs):
        passwd = cls.get_passwd(path)
        _password = passwd.get(username, None)
        if not _password: return False
        return cls.verify(_password, password)

    @classmethod
    def get_passwd(cls, path, cache = True):
        path = os.path.expanduser(path)
        path = os.path.abspath(path)
        path = os.path.normpath(path)

        if not hasattr(cls, "_pwcache"): cls._pwcache = dict()

        result = cls._pwcache.get(path, None) if hasattr(cls, "_pwcache") else None
        if cache and not result == None: return result

        htpasswd = dict()
        contents = cls.get_file(path, cache = cache, encoding = "utf-8")
        for line in contents.split("\n"):
            line = line.strip()
            if not line: continue
            username, password = line.split(":", 1)
            htpasswd[username] = password

        if cache: cls._pwcache[path] = htpasswd
        return htpasswd

    def auth_i(self, username, password, *args, **kwargs):
        return self.__class__.auth(
            username,
            password,
            path = self.path,
            *args,
            **kwargs
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import sys
import time
import socket
import weakref

from . import config
from . import errors
from . import legacy
from . import transport

from . import asynchronous

asyncio = asynchronous.get_asyncio() if asynchronous.is_neo() else None
BaseLoop = asyncio.AbstractEventLoop if asyncio else object

class CompatLoop(BaseLoop):
    """
    Top level compatibility class that adds compatibility support
    for the asyncio event loop strategy.

    This is required to be able to access netius event loop on a
    asyncio like manner.

    :see: https://docs.python.org/3/library/asyncio-eventloop.html
    """

    def __init__(self, loop):
        self._loop = weakref.proxy(loop)
        self._loop_ref = weakref.ref(loop)
        self._task_factory = asynchronous.Task
        self._executor = asynchronous.ThreadPoolExecutor(loop)
        self._handler = self._default_handler

    def __getattr__(self, name):
        if hasattr(self._loop, name):
            return getattr(self._loop, name)
        raise AttributeError("'%s' not found" % name)

    def time(self):
        return time.time()

    def call_soon(self, callback, *args):
        return self._call_delay(callback, args, immediately = True)

    def call_soon_threadsafe(self, callback, *args):
        return self._call_delay(callback, args, immediately = True, safe = True)

    def call_at(self, when, callback, *args):
        delay = when - self.time()
        return self._call_delay(callback, args, timeout = delay)

    def call_later(self, delay, callback, *args):
        """
        Calls the provided callback with the provided parameters after
        the defined delay (in seconds), should ensure proper sleep operation.

        :type delay: float
        :param delay: The delay in seconds after which the callback is going
        to be called with the provided arguments.
        :type callback: Function
        :param callback: The function to be called after the provided delay.
        :rtype: Handle
        :return: The handle object to the operation, that may be used to cancel it.
        """

        return self._call_delay(callback, args, timeout = delay)

    def create_future(self):
        return self._loop.build_future()

    def create_task(self, coroutine):
        future = self._loop.ensure(coroutine)
        task = self._task_factory(future)
        return task

    def create_server(self, *args, **kwargs):
        coroutine = self._create_server(*args, **kwargs)
        return asynchronous.coroutine_return(coroutine)

    def create_connection(self, *args, **kwargs):
        coroutine = self._create_connection(*args, **kwargs)
        return asynchronous.coroutine_return(coroutine)

    def create_datagram_endpoint(self, *args, **kwargs):
        coroutine = self._create_datagram_endpoint(*args, **kwargs)
        return asynchronous.coroutine_return(coroutine)

    def getaddrinfo(self, *args, **kwargs):
        coroutine = self._getaddrinfo(*args, **kwargs)
        return asynchronous.coroutine_return(coroutine)

    def getnameinfo(self, *args, **kwargs):
        coroutine = self._getnameinfo(*args, **kwargs)
        return asynchronous.coroutine_return(coroutine)

    def run_until_complete(self, future):
        self._set_current_task(future)
        try: return self._loop.run_coroutine(future)
        finally: self._unset_current_task()

    def run_forever(self):
        return self._loop.run_forever()

    def run_in_executor(self, *args, **kwargs):
        coroutine = self._run_in_executor(*args, **kwargs)
        return asynchronous.coroutine_return(coroutine)

    def stop(self):
        self._loop.pause()

    def close(self):
        self._loop.close()

    def get_exception_handler(self):
        return self._handler

    def set_exception_handler(self, handler):
        self._handler = handler

    def default_exception_handler(self, context):
        return self._default_handler(context)

    def call_exception_handler(self, context):
        if not self._handler: return
        return self._handler(context)

    def get_debug(self):
        return self._loop.is_debug()

    def set_debug(self, enabled):
        pass

    def set_default_executor(self, executor):
        self._executor = executor

    def get_task_factory(self):
        return self._task_factory

    def set_task_factory(self, factory):
        self._task_factory = factory

    def is_running(self):
        return self._loop.is_running()

    def is_closed(self):
        return self._loop.is_stopped()

    def _getaddrinfo(
        self,
        host,
        port,
        family = 0,
        type = 0,
        proto = 0,
        flags = 0
    ):
        future = self.create_future()
        result = socket.getaddrinfo(
            host,
            port,
            family,
            type,
            proto,
            flags = flags
        )
        self._loop.delay(lambda: future.set_result(result), immediately = True)
        yield future

    def _getnameinfo(self, sockaddr, flags = 0):
        raise errors.NotImplemented("Missing implementation")

    def _run_in_executor(self, executor, func, *args):
        executor = executor or self._executor
        future = executor.submit(func, *args)
        yield future

    def _create_connection(
        self,
        protocol_factory,
        host = None,
        port = None,
        ssl = None,
        family = 0,
        proto = 0,
        flags = 0,
        sock = None,
        local_addr = None,
        server_hostname = None,
        *args,
        **kwargs
    ):
        family = family or socket.AF_INET
        proto = proto or socket.SOCK_STREAM

        future = self.create_future()

        def on_complete(connection, success):
            if success: on_connect(connection)
            else: on_error(connection)

        def on_connect(connection):
            protocol = protocol_factory()
            _transport = transport.TransportStream(self, connection)
            _transport._set_compat(protocol)
            future.set_result((_transport, protocol))

        def on_error(connection):
            future.set_exception(
                errors.RuntimeError("Connection issue")
            )

        self._loop.connect(
            host,
            port,
            ssl = ssl,
            family = family,
            callback = on_complete
        )

        yield future

    def _create_datagram_endpoint(
        self,
        protocol_factory,
        local_addr = None,
        remote_addr = None,
        family = 0,
        proto = 0,
        flags = 0,
        reuse_address = None,
        reuse_port = None,
        allow_broadcast = None,
        sock = None,
        *args,
        **kwargs
    ):
        family = family or socket.AF_INET
        proto = proto or socket.SOCK_DGRAM

        future = self.create_future()

        def on_complete(connection, success):
            if success: on_connect(connection)
            else: on_error(connection)

        def on_connect(connection):
            protocol = protocol_factory()
            _transport = transport.TransportDatagram(self, connection)
            _transport._set_compat(protocol)
            future.set_result((_transport, protocol))

        def on_error(connection):
            future.set_exception(
                errors.RuntimeError("Connection issue")
            )

        connection = self._loop.datagram(
            family = family,
            type = proto,
            local_host = local_addr[0] if local_addr else None,
            local_port = local_addr[1] if local_addr else None,
            remote_host = remote_addr[0] if remote_addr else None,
            remote_port = remote_addr[1] if remote_addr else None
        )

        self._loop.delay(lambda: on_complete(connection, True))
        yield future

    def _set_current_task(self, task):
        asyncio = asynchronous.get_asyncio()
        if not asyncio: return
        asyncio.Task._current_tasks[self] = task

    def _unset_current_task(self):
        asyncio = asynchronous.get_asyncio()
        if not asyncio: return
        asyncio.Task._current_tasks.pop(self, None)

    def _call_delay(
        self,
        callback,
        args,
        timeout = None,
        immediately = False,
        verify = False,
        safe = False
    ):
        # creates the callable to be called after the timeout, note the
        # clojure around the "normal" arguments (allows proper propagation)
        callable = lambda: callback(*args)

        # schedules the delay call of the created callable according to
        # the provided set of options expected by the delay operation the
        # callback tuple is returned so that a proper handle may be created
        callable_t = self._loop.delay(
            callable,
            timeout = timeout,
            immediately = immediately,
            verify = verify,
            safe = safe
        )

        # creates the handle to control the operation and then returns the
        # object to the caller method, allowing operation cancellation
        handle = asynchronous.Handle(callable_t = callable_t)
        return handle

    def _sleep(self, timeout, future = None):
        # verifies if a future variable is meant to be re-used
        # or if instead a new one should be created for the new
        # sleep operation to be executed
        future = future or self.create_future()

        # creates the callable that is going to be used to set
        # the final value of the future variable
        callable = lambda: future.set_result(timeout)

        # delays the execution of the callable so that it is executed
        # after the requested amount of timeout, note that the resolution
        # of the event loop will condition the precision of the timeout
        future._loop.call_later(timeout, callable)
        return future

    def _default_handler(self, context):
        message = context.pop("message", None)
        sys.stderr.write("%s\n" % message)
        for key, value in legacy.iteritems(context):
            sys.stderr.write("%s: %s\n" % (key, value))

    @property
    def _thread_id(self):
        return self._loop.tid

def is_compat():
    """
    Determines if the compatibility mode for the netius
    event loop is required.

    Under this mode the event loop for netius tries to emulate
    the behaviour of the asyncio event loop so that it may
    be used with 3rd party protocol classes (not compliant
    with the netius protocol).

    :rtype: bool
    :return: If the netius infra-structure should run under
    the compatibility mode.
    """

    compat = config.conf("COMPAT", False, cast = bool)
    compat |= is_asyncio()
    return compat and asynchronous.is_neo()

def is_asyncio():
    """
    Checks if the asyncio mode of execution (external event
    loop) is the required approach under the current runtime.

    If that's the case the netius event loop is not going to
    be used and the asyncio one is going to be used instead.

    :rtype: bool
    :return: If the asyncio event loop model is enabled and
    proper library support available.
    """

    asyncio = config.conf("ASYNCIO", False, cast = bool)
    return asyncio and asynchronous.is_asynclib()

def build_datagram(*args, **kwargs):
    if is_compat(): return _build_datagram_compat(*args, **kwargs)
    else: return _build_datagram_native(*args, **kwargs)

def connect_stream(*args, **kwargs):
    if is_compat(): return _connect_stream_compat(*args, **kwargs)
    else: return _connect_stream_native(*args, **kwargs)

def _build_datagram_native(
    protocol_factory,
    family = socket.AF_INET,
    type = socket.SOCK_DGRAM,
    remote_host = None,
    remote_port = None,
    callback = None,
    loop = None,
    *args,
    **kwargs
):
    from . import common

    loop = loop or common.get_loop()

    protocol = protocol_factory()
    has_loop_set = hasattr(protocol, "loop_set")
    if has_loop_set: protocol.loop_set(loop)

    def on_ready():
        loop.datagram(
            family = family,
            type = type,
            remote_host = remote_host,
            remote_port = remote_port,
            callback = on_complete
        )

    def on_complete(connection, success):
        if success: on_connect(connection)
        else: on_error(connection)

    def on_connect(connection):
        _transport = transport.TransportDatagram(loop, connection)
        _transport._set_compat(protocol)
        if not callback: return
        callback((_transport, protocol))

    def on_error(connection):
        protocol.close()

    loop.delay(on_ready)

    return loop

def _build_datagram_compat(
    protocol_factory,
    family = socket.AF_INET,
    type = socket.SOCK_DGRAM,
    remote_host = None,
    remote_port = None,
    callback = None,
    loop = None,
    *args,
    **kwargs
):
    from . import common

    loop = loop or common.get_loop()

    protocol = protocol_factory()
    has_loop_set = hasattr(protocol, "loop_set")
    if has_loop_set: protocol.loop_set(loop)

    def build_protocol():
        return protocol

    def on_connect(future):
        if future.cancelled() or future.exception():
            protocol.close()
        else:
            result = future.result()
            callback and callback(result)

    remote_addr = (remote_host, remote_port) if\
        remote_host and remote_port else kwargs.pop("remote_addr", None)

    connect = loop.create_datagram_endpoint(
        build_protocol,
        family = family,
        remote_addr = remote_addr,
        *args,
        **kwargs
    )

    future = loop.create_task(connect)
    future.add_done_callback(on_connect)

    return loop

def _connect_stream_native(
    protocol_factory,
    host,
    port,
    ssl = False,
    key_file = None,
    cer_file = None,
    ca_file = None,
    ca_root = True,
    ssl_verify = False,
    family = socket.AF_INET,
    type = socket.SOCK_STREAM,
    callback = None,
    loop = None,
    *args,
    **kwargs
):
    from . import common

    loop = loop or common.get_loop()

    protocol = protocol_factory()
    has_loop_set = hasattr(protocol, "loop_set")
    if has_loop_set: protocol.loop_set(loop)

    def on_ready():
        loop.connect(
            host,
            port,
            ssl = ssl,
            key_file = key_file,
            cer_file = cer_file,
            ca_file = ca_file,
            ca_root = ca_root,
            ssl_verify = ssl_verify,
            family = family,
            type = type,
            callback = on_complete
        )

    def on_complete(connection, success):
        if success: on_connect(connection)
        else: on_error(connection)

    def on_connect(connection):
        _transport = transport.TransportStream(loop, connection)
        _transport._set_compat(protocol)
        if not callback: return
        callback((_transport, protocol))

    def on_error(connection):
        protocol.close()

    loop.delay(on_ready)

    return loop

def _connect_stream_compat(
    protocol_factory,
    host,
    port,
    ssl = False,
    key_file = None,
    cer_file = None,
    ca_file = None,
    ca_root = True,
    ssl_verify = False,
    family = socket.AF_INET,
    type = socket.SOCK_STREAM,
    callback = None,
    loop = None,
    *args,
    **kwargs
):
    from . import common

    loop = loop or common.get_loop()

    protocol = protocol_factory()
    has_loop_set = hasattr(protocol, "loop_set")
    if has_loop_set: protocol.loop_set(loop)

    def build_protocol():
        return protocol

    def on_connect(future):
        if future.cancelled() or future.exception():
            protocol.close()
        else:
            result = future.result()
            callback and callback(result)

    if ssl and cer_file and key_file:
        import ssl as _ssl
        ssl_context = _ssl.SSLContext()
        ssl_context.load_cert_chain(cer_file, keyfile = key_file)
        ssl = ssl_context

    connect = loop.create_connection(
        build_protocol,
        host = host,
        port = port,
        ssl = ssl,
        family = family,
        *args,
        **kwargs
    )

    future = loop.create_task(connect)
    future.add_done_callback(on_connect)

    return loop



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import uuid

class NetiusError(Exception):
    """
    The top level base error to be used in the
    netius infra-structure.

    Note that this class inherits from the runtime
    error meaning that all the errors are runtime.
    """

    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args)
        message = args[0] if len(args) > 0 else ""
        kwargs["message"] = kwargs.get("message", message)
        self.kwargs = kwargs
        self._uid = None

    def get_kwarg(self, name, default = None):
        return self.kwargs.get(name, default)

    @property
    def uid(self):
        if self._uid: return self._uid
        self._uid = uuid.uuid4()
        return self._uid

class RuntimeError(NetiusError):
    """
    Error to be used for situations where an exception
    is raised during a typical runtime situation.

    This error class is meant to be used as the parent
    class in every exception raised during normal execution.
    """

    pass

class StopError(RuntimeError):
    """
    Error to be used for situations where a stop
    intention is meant to be raised to upper layers.

    This error represent an operation and not a real
    error and should be used as such.
    """

    pass

class PauseError(RuntimeError):
    """
    Error to be used for situations where a pause
    intention is meant to be raised to upper layers.

    This error represent an operation and not a real
    error and should be used as such.
    """

    pass

class WakeupError(RuntimeError):
    """
    Error used to send a wakeup intent from one context
    or thread to another.

    This is especially useful on the context of signal
    handling where an interruption may happen at any time.
    """

    pass

class DataError(RuntimeError):
    """
    Error to be used for situations where the
    data that has been received/sent is invalid.

    This error may be used for situations where
    the data in the buffer is not sufficient for
    parsing the values.
    """

    pass

class ParserError(RuntimeError):
    """
    Error caused by a malformed data that invalidated
    the possibility to parse it.

    This error should only be used under a parser infra-
    structure and never outside it.
    """

    pass

class GeneratorError(RuntimeError):
    """
    Error generated by a problem in the generation of
    and encoded data (reverse of parser error).

    This error should be raise only in a generator of
    an encoded stream buffer.
    """

    pass

class SecurityError(RuntimeError):
    """
    Error caused by a failed security verification this
    errors should be properly audited in order to avoid
    extra problems that may arise from them.

    This kind of problems is considered to be runtime
    as they should not be related with programming.
    """

    pass

class NotImplemented(RuntimeError):
    """
    Error caused by the non implementation of a certain
    method/feature at a certain level. This may mean that
    the wrong level of abstraction is being called or a
    certain feature is pending development.

    This kind of problems is considered to be development
    as they may be related with programming.
    """

    pass

class AssertionError(RuntimeError):
    """
    Error raised for failure to meet any pre-condition or
    assertion for a certain data set.
    """

    pass



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import threading

from . import legacy
from . import observer

class Agent(observer.Observable):
    """
    Top level class for the entry point classes of the multiple
    client and server protocol implementations.

    These classes should contain a series of utilities that facilitate
    the interaction with the Protocol, Event Loop and Transport
    objects (with the end developer in mind).

    Most of the interaction for a simple protocol should be implemented
    using static or class methods, avoiding internal object state and
    instantiation of the concrete Agent class.

    For complex protocols instantiation may be useful to provided extra
    flexibility and context for abstract operations.
    """

    @classmethod
    def cleanup_s(cls):
        pass

    def cleanup(self, destroy = True):
        if destroy: self.destroy()

    def destroy(self):
        observer.Observable.destroy(self)

class ClientAgent(Agent):

    _clients = dict()
    """ The global static clients map meant to be reused by the
    various static clients that may be created, this client
    may leak creating blocking threads that will prevent the
    system from exiting correctly, in order to prevent that
    the cleanup method should be called """

    @classmethod
    def cleanup_s(cls):
        super(ClientAgent, cls).cleanup_s()
        for client in legacy.itervalues(cls._clients):
            client.close()
        cls._clients.clear()

    @classmethod
    def get_client_s(cls, *args, **kwargs):
        tid = threading.current_thread().ident
        client = cls._clients.get(tid, None)
        if client: return client
        client = cls(*args, **kwargs)
        cls._clients[tid] = client
        return client

class ServerAgent(Agent):
    pass



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import time
import struct

import netius.common

HANDSHAKE_STATE = 1

NORMAL_STATE = 2

CHOKED = 1

UNCHOKED = 2

ALIVE_TIMEOUT = 45.0
""" The timeout that is going to be used in the operation of
keep alive the connection that are active, any connections that
does not send a message with an interval less than this timeout
is going to be disconnected """

SPEED_LIMIT = 10240
""" The minimum download speed limit from which a connection will
be disconnected if does not fulfill such limit, this is going to
optimize the connection in competition for blocks """

BLOCK_SIZE = 16384
""" The typical size of block that is going to be retrieved
using the current torrent infra-structure, this value conditions
most of the torrent operations and should be defined carefully """

class TorrentConnection(netius.Connection):

    def __init__(self, max_requests = 50, *args, **kwargs):
        netius.Connection.__init__(self, *args, **kwargs)
        self.parser = None
        self.max_requests = max_requests
        self.pend_requests = 0
        self.task = None
        self.peer_id = None
        self.bitfield = b""
        self.state = HANDSHAKE_STATE
        self.choked = CHOKED
        self.start = time.time()
        self.messages = 0
        self.downloaded = 0
        self.requests = []

    def open(self, *args, **kwargs):
        netius.Connection.open(self, *args, **kwargs)
        if not self.is_open(): return
        self.parser = netius.common.TorrentParser(self)
        self.bind("close", self.on_close)
        self.parser.bind("on_handshake", self.on_handshake)
        self.parser.bind("on_message", self.on_message)
        self.is_alive(timeout = ALIVE_TIMEOUT, schedule = True)

    def close(self, *args, **kwargs):
        netius.Connection.close(self, *args, **kwargs)
        if not self.is_closed(): return
        if self.parser: self.parser.destroy()

    def on_close(self, connection):
        self.release()

    def on_handshake(self, protocol, reserved, info_hash, peer_id):
        self.peer_id = peer_id
        self.state = NORMAL_STATE
        self.interested()
        self.unchoke()

    def on_message(self, length, type, data):
        self.handle(type, data)
        self.messages += 1

    def parse(self, data):
        self.parser.parse(data)

    def handle(self, type, data):
        # constructs the name of the method that is going to be called
        # for the handle of the message from the provided type and verifies
        # if the method exists under the current instance
        method_name = "%s_t" % type
        if not hasattr(self, method_name): return

        # tries to retrieve the method for the current state in iteration
        # and then calls the retrieve method with (handler method)
        method = getattr(self, method_name)
        method(data)

    def bitfield_t(self, data):
        bitfield = netius.common.string_to_bits(data)
        self.bitfield = [True if value == "1" else False for value in bitfield]

    def choke_t(self, data):
        if self.choked == CHOKED: return
        self.choked = CHOKED
        self.release()
        self.trigger("choked", self)

    def unchoke_t(self, data):
        if self.choked == UNCHOKED: return
        self.choked = UNCHOKED
        self.reset()
        self.next()
        self.trigger("unchoked", self)

    def piece_t(self, data):
        block = struct.unpack("!LL", data[:8])
        index, begin = block
        data = data[8:]
        self.task.set_data(data, index, begin)
        self.downloaded += len(data)
        self.remove_request(block)
        self.next()
        self.trigger("piece", self, data, index, begin)

    def port_t(self, data):
        port, = struct.unpack("!H", data[:8])
        self.task.set_dht(self.address, port)

    def next(self, count = None):
        if not self.choked == UNCHOKED: return
        if count == None: count = self.max_requests - self.pend_requests
        for _index in range(count):
            block = self.task.pop_block(self.bitfield)
            if not block: return
            index, begin, length = block
            self.request(index, begin = begin, length = length)
            block_t = (index, begin)
            self.add_request(block_t)

    def add_request(self, block):
        self.requests.append(block)
        self.pend_requests += 1

    def remove_request(self, block):
        if not block in self.requests: return
        self.requests.remove(block)
        self.pend_requests -= 1

    def reset(self):
        del self.requests[:]
        self.pend_requests = 0

    def release(self):
        for index, begin in self.requests:
            self.task.push_block(index, begin)
        self.reset()

    def handshake(self):
        data = struct.pack(
            "!B19sQ20s20s",
            19,
            b"BitTorrent protocol",
            1,
            self.task.info_hash,
            netius.legacy.bytes(self.task.owner.peer_id)
        )
        data and self.send(data)

    def keep_alive(self):
        data = struct.pack("!L", 0)
        data and self.send(data)

    def choke(self):
        data = struct.pack("!LB", 1, 0)
        data and self.send(data)

    def unchoke(self):
        data = struct.pack("!LB", 1, 1)
        data and self.send(data)

    def interested(self):
        data = struct.pack("!LB", 1, 2)
        data and self.send(data)

    def not_interested(self):
        data = struct.pack("!LB", 1, 3)
        data and self.send(data)

    def have(self, index):
        data = struct.pack("!LBL", 5, 4, index)
        data and self.send(data)

    def request(self, index, begin = 0, length = BLOCK_SIZE):
        data = struct.pack("!LBLLL", 13, 6, index, begin, length)
        data and self.send(data)

    def is_alive(self, timeout = ALIVE_TIMEOUT, schedule = False):
        messages = self.messages
        downloaded = self.downloaded

        def clojure():
            if not self.is_open(): return
            delta = self.downloaded - downloaded
            rate = float(delta) / float(timeout)
            if self.messages == messages: self.close(flush = True); return
            if rate < SPEED_LIMIT: self.close(flush = True); return
            callable = self.is_alive()
            self.owner.delay(callable, timeout)

        if schedule: self.owner.delay(clojure, timeout)
        return clojure

class TorrentClient(netius.StreamClient):
    """
    Implementation of the torrent protocol, able to download
    and seed files across a peer to peer mesh network.

    The client provides a series of top level methods that
    provide the main interface with the system.

    The current implementation support both a torrent file
    (using trackers) strategy and also a DHT (distributed
    has table) strategy for completely decentralized usage.

    :see: http://www.bittorrent.org/beps/bep_0003.html
    """

    def peer(self, task, host, port, ssl = False, connection = None):
        connection = connection or self.acquire_c(host, port, ssl = ssl)
        connection.task = task
        return connection

    def on_connect(self, connection):
        netius.StreamClient.on_connect(self, connection)
        self.trigger("connect", self, connection)

    def on_acquire(self, connection):
        netius.StreamClient.on_acquire(self, connection)
        self.trigger("acquire", self, connection)
        connection.handshake()

    def on_data(self, connection, data):
        netius.StreamClient.on_data(self, connection, data)
        connection.parse(data)

    def build_connection(self, socket, address, ssl = False):
        return TorrentConnection(
            owner = self,
            socket = socket,
            address = address,
            ssl = ssl
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import time

import netius

from .base import Middleware

class FloodMiddleware(Middleware):
    """
    Simple middleware implementation for avoiding flooding
    of connection creation from a certain address.
    """

    def __init__(self, owner, conns_per_min = 600, whitelist = None):
        Middleware.__init__(self, owner)
        self.blacklist = conns_per_min
        self.whitelist = whitelist or []

    def start(self):
        Middleware.start(self)
        self.conns_per_min = netius.conf("CONNS_PER_MIN", self.conns_per_min, cast = int)
        self.whitelist = netius.conf("WHITELIST", self.whitelist, cast = list)
        self.blacklist = []
        self.conn_map = dict()
        self.minute = int(time.time() // 60)
        self.owner.bind("connection_c", self.on_connection_c)

    def stop(self):
        Middleware.stop(self)
        self.owner.unbind("connection_c", self.on_connection_c)

    def on_connection_c(self, owner, connection):
        host = connection.address[0]
        self._update_flood(host)
        if not host in self.blacklist and not "*" in self.blacklist: return
        if host in self.whitelist: return
        self.owner.warning("Connection from '%s' dropped (flooding avoidance)" % host)
        connection.close()

    def _update_flood(self, host):
        minute = int(time.time() // 60)
        if minute == self.minute: self.conn_map.clear()
        self.minute = minute
        count = self.conn_map.get(host, 0)
        count += 1
        self.conn_map[host] = count
        if count > self.conns_per_min: self.blacklist.append(host)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import uuid
import base64
import hashlib

import netius.common

class WSProtocol(netius.StreamProtocol):
    """
    Abstract WebSockets protocol to be used for real-time bidirectional
    communication on top of the HTTP protocol.

    :see: https://tools.ietf.org/html/rfc6455
    """

    MAGIC_VALUE = "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"
    """ The magic value used by the websocket protocol as part
    of the key generation process in the handshake """

    @classmethod
    def _key(cls, size = 16):
        seed = str(uuid.uuid4())
        seed = netius.legacy.bytes(seed)[:size]
        seed = base64.b64encode(seed)
        return netius.legacy.str(seed)

    def __init__(self, *args, **kwargs):
        netius.StreamProtocol.__init__(self, *args, **kwargs)
        self.host = None
        self.port = None
        self.ssl = False
        self.path = None
        self.key = None
        self.version = None
        self.code = None
        self.code_s = None
        self.handshake = False
        self.buffer_l = []
        self.headers = {}

    def connection_made(self, transport):
        netius.StreamProtocol.connection_made(self, transport)
        data = "GET %s HTTP/1.1\r\n" % self.path +\
            "Upgrade: websocket\r\n" +\
            "Connection: Upgrade\r\n" +\
            "Host: %s\r\n" % self.host +\
            "Origin: http://%s\r\n" % self.host +\
            "Sec-WebSocket-Key: %s\r\n" % self.key +\
            "Sec-WebSocket-Version: 13\r\n\r\n"
        self.send(data)

    def on_data(self, data):
        netius.StreamProtocol.on_data(self, data)

        # iterates while there's still data pending to be parsed from the
        # current message received using the HTTP or WS protocols
        while data:
            if self.handshake:
                # retrieves the current (pending) buffer of data for the
                # protocol and tries to run the decoder of websockets
                # frame on the complete set of data pending in case there's
                # a problem the (pending) data is added to the buffer
                buffer = self.get_buffer()
                data = buffer + data
                try: decoded, data = netius.common.decode_ws(data)
                except netius.DataError: self.add_buffer(data); break

                # calls the callback method in the protocol notifying
                # it about the new (decoded) data that has been received
                self.receive_ws(decoded)

                # calls the proper callback handler for the data
                # that has been received with the decoded data
                self.on_data_ws(decoded)

            else:
                # adds the current data to the internal protocol
                # buffer to be processed latter, by the handshake
                self.add_buffer(data)

                # tries to run the handshake operation for the
                # current protocol in case it fails due to an
                # handshake error must delay the execution to the
                # next iteration (not enough data)
                try: self.do_handshake()
                except netius.DataError: return

                # validates (and computes) the accept key value according
                # to the provided value, in case there's an error an exception
                # should be raised (finishing the protocol)
                self.validate_key()

                # calls the on handshake event handler for the current
                # protocol to notify the current object
                self.on_handshake()

                # retrieves the current buffer value as the data that is still
                # pending to be parsed from the current protocol, this is
                # required so that the complete client buffer is flushed
                data = self.get_buffer()

    def on_data_ws(self, data):
        self.trigger("message", self, data)

    def on_handshake(self):
        self.trigger("handshake", self)

    def connect_ws(self, url, callback = None, loop = None):
        cls = self.__class__

        parsed = netius.legacy.urlparse(url)
        self.ssl = parsed.scheme == "wss"
        self.host = parsed.hostname
        self.port = parsed.port or (self.ssl and 443 or 80)
        self.path = parsed.path or "/"

        loop = netius.connect_stream(
            lambda: self,
            host = self.host,
            port = self.port,
            ssl = self.ssl,
            loop = loop
        )

        self.key = cls._key()

        if callback: self.bind("handshake", callback, oneshot = True)

        return loop, self

    def send_ws(self, data, callback = None):
        encoded = netius.common.encode_ws(data, mask = True)
        return self.send(encoded, callback = callback)

    def receive_ws(self, decoded):
        pass

    def add_buffer(self, data):
        self.buffer_l.append(data)

    def get_buffer(self, delete = True):
        if not self.buffer_l: return b""
        buffer = b"".join(self.buffer_l)
        if delete: del self.buffer_l[:]
        return buffer

    def do_handshake(self):
        if self.handshake:
            raise netius.NetiusError("Handshake already done")

        buffer = b"".join(self.buffer_l)
        end_index = buffer.find(b"\r\n\r\n")
        if end_index == -1: raise netius.DataError("Missing data for handshake")

        data = buffer[:end_index + 4]
        remaining = buffer[end_index + 4:]

        lines = data.split(b"\r\n")
        for line in lines[1:]:
            values = line.split(b":", 1)
            values_l = len(values)
            if not values_l == 2: continue

            key, value = values
            key = key.strip()
            key = netius.legacy.str(key)
            value = value.strip()
            value = netius.legacy.str(value)
            self.headers[key] = value

        first = lines[0]
        first = netius.legacy.str(first)
        self.version, self.code, self.code_s = first.split(" ", 2)

        del self.buffer_l[:]
        self.handshake = True

        if remaining: self.add_buffer(remaining)

    def validate_key(self):
        accept_key = self.headers.get("Sec-WebSocket-Accept", None)
        if not accept_key:
            raise netius.NetiusError("No accept key found in headers")

        value = netius.legacy.bytes(self.key + WSProtocol.MAGIC_VALUE)
        hash = hashlib.sha1(value)
        hash_digest = hash.digest()
        _accept_key = base64.b64encode(hash_digest)
        _accept_key = netius.legacy.str(_accept_key)

        if not _accept_key == accept_key:
            raise netius.SecurityError("Invalid accept key provided")

class WSClient(netius.ClientAgent):

    protocol = WSProtocol

    @classmethod
    def connect_ws_s(cls, url, callback = None, loop = None):
        protocol = cls.protocol()
        return protocol.connect_ws(url, callback = callback, loop = loop)

if __name__ == "__main__":
    def on_handshake(protocol):
        protocol.send_ws("Hello World")

    def on_message(protocol, data):
        print(data)
        protocol.close()

    def on_close(protocol):
        netius.compat_loop(loop).stop()

    url = netius.conf("WS_URL", "ws://echo.websocket.org/")

    loop, protocol = WSClient.connect_ws_s(url)

    protocol.bind("handshake", on_handshake)
    protocol.bind("message", on_message)
    protocol.bind("close", on_close)

    loop.run_forever()
    loop.close()
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import hashlib
import binascii

import netius

class Auth(object):
    """
    The top level base authentication handler, should define
    and implement generic authentication methods.

    The proper exceptions must be raised when the implementation
    at this abstraction level is insufficient or insecure.
    """

    def __init__(self, *args, **kwargs):
        object.__init__(self, *args, **kwargs)
        self.auth = self.auth_i
        self.auth_assert = self.auth_assert_i

    @classmethod
    def auth(cls, *args, **kwargs):
        raise netius.NotImplemented("Missing implementation")

    @classmethod
    def auth_assert(cls, *args, **kwargs):
        result = cls.auth(*args, **kwargs)
        if not result: raise netius.SecurityError("Invalid authentication")

    @classmethod
    def verify(cls, encoded, decoded):
        type, salt, digest, plain = cls.unpack(encoded)
        if plain: return encoded == decoded
        if salt: decoded += salt
        type = type.lower()
        decoded = netius.legacy.bytes(decoded)
        hash = hashlib.new(type, decoded)
        _digest = hash.hexdigest()
        return _digest == digest

    @classmethod
    def generate(cls, password, type = "sha256", salt = "netius"):
        if type == "plain" : return password
        if salt: password += salt
        password = netius.legacy.bytes(password)
        hash = hashlib.new(type, password)
        digest = hash.hexdigest()
        if not salt: return "%s:%s" % (type, digest)
        salt = netius.legacy.bytes(salt)
        salt = binascii.hexlify(salt)
        salt = netius.legacy.str(salt)
        return "%s:%s:%s" % (type, salt, digest)

    @classmethod
    def unpack(cls, password):
        count = password.count(":")
        if count == 2: type, salt, digest = password.split(":")
        elif count == 1: type, digest = password.split(":"); salt = None
        else: plain = password; type = "plain"; salt = None; digest = None
        if not type == "plain": plain = None
        if salt: salt = netius.legacy.bytes(salt)
        if salt: salt = binascii.unhexlify(salt)
        if salt: salt = netius.legacy.str(salt)
        return (type, salt, digest, plain)

    @classmethod
    def get_file(cls, path, cache = False, encoding = None):
        """
        Retrieves the (file) contents for the file located "under"
        the provided path, these contents are returned as a normal
        string based byte buffer.

        In case the cache flag is set these contents are store in
        memory so that they be latter retrieved much faster.

        If the optional encoding value is provide the final value is
        decoded according to the defined encoding.

        :type path: String
        :param path: The path as string to the file for which the
        contents are going to be retrieved.
        :type cache: bool
        :param cache: If the contents should be stored in memory and
        associated with the current path for latter access.
        :type encoding: String
        :param encoding: The string that identifies the encoding that
        is going to be used for the decoding of the final string value.
        :rtype: String
        :return: The contents (as a string) of the file located under
        the provided path (from the file system).
        """

        # runs the complete set of normalization processes for the path so
        # that the final path to be used in retrieval is canonical, providing
        # a better mechanisms for both loading and cache processes
        path = os.path.expanduser(path)
        path = os.path.abspath(path)
        path = os.path.normpath(path)

        # verifies if the cache attribute already exists under the current class
        # and in case it does not creates the initial cache dictionary
        if not hasattr(cls, "_cache"): cls._cache = dict()

        # tries to retrieve the contents of the file using a caches approach
        # and returns such value in case the cache flag is enabled
        result = cls._cache.get(path, None)
        if cache and not result == None: return result

        # as the cache retrieval has not been successful there's a need to
        # load the file from the secondary storage (file system)
        file = open(path, "rb")
        try: contents = file.read()
        finally: file.close()

        # in case an encoding value has been passed the contents must be properly
        # decoded so that the "final" contents string is defined
        if encoding: contents = contents.decode(encoding)

        # verifies if the cache mode/flag is enabled and if that's the case
        # store the complete file contents in memory under the dictionary
        if cache: cls._cache[path] = contents
        return contents

    @classmethod
    def is_simple(cls):
        return False

    def auth_i(self, *args, **kwargs):
        return self.__class__.auth(*args, **kwargs)

    def auth_assert_i(self, *args, **kwargs):
        result = self.auth_i(*args, **kwargs)
        if not result: raise netius.SecurityError("Invalid authentication")

    def is_simple_i(self):
        return self.__class__.is_simple()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import re

import netius

from . import parser

SEPARATOR_REGEX = re.compile(r" |\-")
""" Regular expression that is going to be used to split
the various response lines between the code and the message
part, it should handle both normal and continuation lines """

class SMTPParser(parser.Parser):

    def __init__(self, owner, store = False):
        parser.Parser.__init__(self, owner)

        self.buffer = []

    def parse(self, data):
        """
        Parses the provided data chunk, changing the current
        state of the parser accordingly and returning the
        number of processed bytes from it.

        :type data: String
        :param data: The string containing the data to be parsed
        in the current parse operation.
        :rtype: int
        :return: The amount of bytes of the data string that have
        been "parsed" in the current parse operation.
        """

        parser.Parser.parse(self, data)

        # retrieves the size of the data that has been sent for parsing
        # and saves it under the size original variable
        size = len(data)
        size_o = size

        # iterates continuously to try to process all that
        # data that has been sent for processing
        while size > 0:
            # retrieves the parsing method for the current
            # state and then runs it retrieving the number
            # of valid parsed bytes in case this value is
            # zero the parsing iteration is broken
            method = self._parse_line
            count = method(data)
            if count == 0: break

            # decrements the size of the data buffer by the
            # size of the parsed bytes and then retrieves the
            # sub part of the data buffer as the new data buffer
            size -= count
            data = data[count:]

        # in case not all of the data has been processed
        # must add it to the buffer so that it may be used
        # latter in the next parsing of the message
        if size > 0: self.buffer.append(data)

        # returns the number of read (processed) bytes of the
        # data that has been sent to the parser
        return size_o - size

    def _parse_line(self, data):
        # tries to find the new line character in the currently received
        # data in case it's not found returns immediately with no data processed
        index = data.find(b"\n")
        if index == -1: return 0

        # adds the partial data (until new line) to the current buffer and
        # then joins it retrieving the current line, then deletes the buffer
        # as it's not longer going to be used
        self.buffer.append(data[:index])
        self.line_s = b"".join(self.buffer).rstrip()
        self.line_s = netius.legacy.str(self.line_s)
        del self.buffer[:]

        # splits the provided line into the code and message parts in case
        # the split is not successful (not enough information) then an extra
        # value is added to the sequence of values for compatibility
        values = SEPARATOR_REGEX.split(self.line_s, 1)
        if not len(values) > 1: values.append("")

        # unpacks the set of values that have just been parsed into the code
        # and the message items as expected by the smtp specification
        code, message = values

        # verifies if the current line is a final line meaning that no more
        # lines are going to be sent as part of this response, a final line
        # is one that does not have the continuation character in it as part
        # of the first separator character in the line
        line_l = len(self.line_s)
        space_index = self.line_s.find(" ")
        token_index = self.line_s.find("-")
        if space_index == -1: space_index = line_l
        if token_index == -1: token_index = line_l
        is_continuation = token_index < space_index
        is_final = not is_continuation

        # triggers the on line event so that the listeners are notified
        # about the end of the parsing of the smtp line and then
        # returns the count of the parsed bytes of the message
        self.trigger("on_line", code, message, is_final = is_final)
        return index + 1



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

from . import file as _file

BUFFER_SIZE = _file.BUFFER_SIZE * 4
""" Re-creates the buffer size of the file buffer so that it
handles more data for each chunk, this is required to avoid
extreme amounts of overhead in the file pool """

class FileAsyncServer(_file.FileServer):
    """
    Simple implementation of a file server that uses the async
    file pool infra-structure.

    This is a test implementation and should never be used for
    production work that required mature and stable codebase.

    Using this kind of server (file pool based) is not recommended
    for system that don't provide some system of event fd (eg: windows)
    as it would provide very slow performance or even stall the
    event loop as no notification occurs on events.
    """

    def on_connection_d(self, connection):
        file = hasattr(connection, "file") and connection.file
        if file: self.fclose(file); connection.file = None
        _file.FileServer.on_connection_d(self, connection)

    def on_stream_d(self, stream):
        file = hasattr(stream, "file") and stream.file
        if file: self.fclose(file); stream.file = None
        _file.FileServer.on_stream_d(self, stream)

    def _file_send(self, connection):
        file = connection.file
        range = connection.range
        is_larger = BUFFER_SIZE > connection.bytes_p
        buffer_s = connection.bytes_p if is_larger else BUFFER_SIZE

        def callback(data, *args, **kwargs):
            if connection.file == None: return
            if isinstance(data, BaseException): return
            data_l = len(data) if data else 0
            connection.bytes_p -= data_l
            is_final = not data or connection.bytes_p == 0
            callback = self._file_finish if is_final else self._file_send
            connection.send_part(
                data,
                final = False,
                callback = callback
            )

        self.fread(file, buffer_s, data = callback)

if __name__ == "__main__":
    import logging

    server = FileAsyncServer(level = logging.INFO)
    server.serve(env = True)
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

BASE_STYLE = """<style>
@charset "utf-8";
@import \"//fonts.googleapis.com/css?family=Open+Sans\";
html {
    height: 100%;
}
body {
    color: #2d2d2d;
    font-family: -apple-system, \"BlinkMacSystemFont\", \"Segoe UI\", \"Roboto\", \"Open Sans\", \"Helvetica\", \"Arial\", sans-serif;
    font-size: 13px;
    line-height: 18px;
    shape-rendering: crispEdges;
}
iframe {
    border-width: 0px;
    height: 100%;
    left: 0px;
    position: absolute;
    top: 0px;
    width: 100%;
}
h1 {
    font-size: 22px;
    font-weight: 600;
    line-height: 26px;
    margin: 14px 0px 14px 0px;
}
a {
    color: #4769cc;
    text-decoration: none;
}
a:hover {
    text-decoration: underline;
}
hr {
    margin: 3px 0px 3px 0px;
}
table {
    font-size: 13px;
    line-height: 18px;
    max-width: 760px;
    table-layout: fixed;
    word-break: break-all;
}
table th {
    font-weight: 600;
}
table th > *, table td > * {
    vertical-align: middle;
}
table th > a {
    color: #2d2d2d;
}
table th > a.selected {
    font-weight: bold;
    text-decoration: underline;
}
table td > svg {
    color: #4d4d4d;
    fill: currentColor;
    margin-right: 6px;
}
.traceback {
    line-height: 26px;
    margin: 8px 0px 8px 0px;
}
@media screen and (max-width: 760px) {
    table th, table td {
        display: none;
    }
    table td:nth-child(1) {
        display: initial;
    }
}
</style>"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import netius.servers

class HelloServer(netius.servers.HTTP2Server):
    """
    Simple Hello (World) HTTP server meant to be used for benchmarks
    and other operations that require a simple in-memory HTTP server.

    Most of the implementation on the server is done in the upper
    layers from which this server class inherits.

    Performance should always be considered critical when changing
    or adding new features to this server implementation.
    """

    def __init__(self, message = "Hello World", *args, **kwargs):
        netius.servers.HTTP2Server.__init__(self, *args, **kwargs)
        self.message = message

    def on_serve(self):
        netius.servers.HTTP2Server.on_serve(self)
        if self.env: self.message = self.get_env("MESSAGE", self.message, cast = str)
        if self.env: self.keep_alive = self.get_env("KEEP_ALIVE", True, cast = bool)
        self.info("Serving '%s' as welcome message ..." % self.message)

    def on_data_http(self, connection, parser):
        netius.servers.HTTP2Server.on_data_http(
            self, connection, parser
        )

        keep_alive = self.keep_alive and parser.keep_alive
        callback = self._hello_keep if keep_alive else self._hello_close
        connection_s = "keep-alive" if keep_alive else "close"
        headers = {
            "Connection" : connection_s,
            "Content-Type" : "text/plain"
        }

        connection.send_response(
            data = self.message,
            headers = headers,
            code = 200,
            code_s = "OK",
            apply = True,
            callback = callback
        )

    def _hello_close(self, connection):
        self.delay(connection.close)

    def _hello_keep(self, connection):
        pass

if __name__ == "__main__":
    import logging
    server = HelloServer(level = logging.INFO)
    server.serve(env = True)
else:
    __path__ = []
